# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import pandas as pd

from xcube.server.api import ApiHandler
from .api import api
from .context import TimeSeriesContext
from .controllers import get_time_series
from ..datasets import PATH_PARAM_DATASET_ID
from ..datasets import PATH_PARAM_VAR_NAME


# noinspection PyPep8Naming
@api.route('/timeseries/{datasetId}/{varName}')
class TimeseriesHandler(ApiHandler[TimeSeriesContext]):
    @api.operation(operation_id='getTimeSeries',
                   summary="Get the time-series for a variable"
                           " and given GeoJSON object.",
                   parameters=[
                       PATH_PARAM_DATASET_ID,
                       PATH_PARAM_VAR_NAME,
                       {
                           "name": "aggMethods",
                           "in": "query",
                           "description": 'Comma-separated list of'
                                          ' aggregation methods.'
                                          ' Valid methods are'
                                          ' "mean", "median", "std",'
                                          ' "min", "max", "count"',
                           "schema": {
                               "type": "string",
                           }
                       },
                       {
                           "name": "startDate",
                           "in": "query",
                           "description": "Start timestamp",
                           "schema": {
                               "type": "string",
                               "format": "datetime"
                           }
                       },
                       {
                           "name": "endDate",
                           "in": "query",
                           "description": "End timestamp",
                           "schema": {
                               "type": "string",
                               "format": "datetime"
                           }
                       },
                       {
                           "name": "tolerance",
                           "in": "query",
                           "description": "Time tolerance in seconds that"
                                          " expands the given time range",
                           "schema": {
                               "type": "string",
                               "default": 1.0,
                           }
                       },
                       {
                           "name": "maxValids",
                           "in": "query",
                           "description": "Maximum number of valid"
                                          " time-series values"
                                          " to be returned.",
                           "schema": {
                               "type": "integer",
                           }
                       },
                   ])
    async def post(self, datasetId: str, varName: str):
        geo_json_object = self.request.json
        agg_methods = self.request.get_query_arg('aggMethods',
                                                 type=str,
                                                 default=None)
        agg_methods = agg_methods.split(',') if agg_methods else None
        start_date = self.request.get_query_arg('startDate',
                                                type=pd.Timestamp,
                                                default=None)
        end_date = self.request.get_query_arg('endDate',
                                              type=pd.Timestamp,
                                              default=None)
        tolerance = self.request.get_query_arg('tolerance',
                                               type=float,
                                               default=1.0)
        max_valids = self.request.get_query_arg('maxValids',
                                                type=int,
                                                default=None)
        result = await self.ctx.run_in_executor(None,
                                                get_time_series,
                                                self.ctx,
                                                datasetId,
                                                varName,
                                                geo_json_object,
                                                agg_methods,
                                                start_date,
                                                end_date,
                                                tolerance,
                                                max_valids)
        self.response.set_header('Content-Type', 'application/json')
        await self.response.finish(dict(result=result))



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
import re
from typing import Mapping, Sequence, Optional, Union
import pyproj
import rfc3339_validator


class CoverageRequest:
    """A representation of a parsed OGC API - Coverages request

    As defined in https://docs.ogc.org/DRAFTS/19-087.html
    (§7.2.4–11.2.1)
    """

    properties: Optional[list[str]]
    bbox: Optional[list[float]]
    bbox_crs: pyproj.CRS
    datetime: Optional[Union[str, tuple[Optional[str], Optional[str]]]]
    subset: Optional[
        dict[str, Union[str, tuple[Optional[str], Optional[str]]]]
    ]
    subset_crs: pyproj.CRS
    scale_factor: Optional[float]
    scale_axes: Optional[dict[str, float]]
    scale_size: Optional[dict[str, float]]
    crs: Optional[pyproj.CRS]

    def __init__(self, query: Mapping[str, Sequence[str]]):
        self.query = query
        self._parse_bbox()
        self.bbox_crs = self._parse_crs('bbox-crs', 'OGC:CRS84')
        self._parse_properties()
        self._parse_datetime()
        self._parse_subset()
        self.subset_crs = self._parse_crs('subset-crs', 'OGC:CRS84')
        self._parse_scale_factor()
        self._parse_scale_axes()
        self._parse_scale_size()
        self.crs = self._parse_crs('crs')

    def _parse_bbox(self):
        self.bbox = None
        if 'bbox' in self.query:
            bbox_spec = self.query['bbox'][0]
            try:
                self.bbox = list(map(float, bbox_spec.split(',')))
            except ValueError:
                raise ValueError(f'Invalid bbox "{bbox_spec}"')
            if len(self.bbox) != 4:
                # TODO: Handle 3D bounding boxes
                raise ValueError(
                    f'Invalid bbox "{bbox_spec}": must have 4 elements'
                )

    def _parse_datetime(self):
        if 'datetime' in self.query:
            datetime_spec = self.query['datetime'][0]
            limits = datetime_spec.split('/')
            if len(limits) > 2:
                raise ValueError(f'Too many parts in datetime {datetime_spec}')
            limits = tuple(None if lt == '..' else lt for lt in limits)
            for limit in limits:
                if (
                    limit is not None
                    and not rfc3339_validator.validate_rfc3339(limit)
                ):
                    raise ValueError(f'Invalid datetime {limit}')
            self.datetime = limits if len(limits) == 2 else limits[0]
        else:
            self.datetime = None

    def _parse_subset(self):
        if 'subset' in self.query:
            self.subset = {}
            subset_spec = self.query['subset'][0]
            for part in subset_spec.split(','):
                # First try matching with quotation marks
                m = re.match('^(.*)[(]"([^")]*)"(?::"(.*)")?[)]$', part)
                if m is None:
                    # If that fails, try without quotation marks
                    m = re.match('^(.*)[(]([^:)]*)(?::(.*))?[)]$', part)
                if m is None:
                    raise ValueError(f'Unrecognized subset specifier "{part}"')
                else:
                    axis, low, high = m.groups()
                if high is None:
                    self.subset[axis] = low
                else:
                    low = None if low == '*' else low
                    high = None if high == '*' else high
                    self.subset[axis] = low, high
        else:
            self.subset = None

    def _parse_properties(self):
        # https://docs.ogc.org/DRAFTS/19-087.html#_parameter_properties
        # TODO: Support * wildcard (Requirement 13E)
        if 'properties' in self.query:
            self.properties = self.query['properties'][0].split(',')
        else:
            self.properties = None

    def _parse_scale_factor(self):
        if 'scale-factor' in self.query:
            scale_factor_str = self.query['scale-factor'][0]
            try:
                self.scale_factor = float(scale_factor_str)
            except ValueError:
                raise ValueError(f'Invalid scale-factor {scale_factor_str}')
        else:
            # We don't default to 1, since (per the standard) an implementation
            # may choose to downscale by default.
            self.scale_factor = None

    def _parse_scale_axes(self):
        self.scale_axes = (
            self._parse_scale_specifier('scale-axes')
            if 'scale-axes' in self.query
            else None
        )

    def _parse_scale_size(self):
        self.scale_size = (
            self._parse_scale_specifier('scale-size')
            if 'scale-size' in self.query
            else None
        )

    def _parse_scale_specifier(self, param_name: str) -> dict[str, float]:
        spec = self.query[param_name][0]
        result = {}
        for part in spec.split(','):
            m = re.match('^(.*)[(]([0-9.]+)[)]$', part)
            if m is None:
                raise ValueError(f'Invalid {param_name} value {part}')
            else:
                result[m.group(1)] = float(m.group(2))
        return result

    def _parse_crs(self, param, default=None):
        specifier = self.query[param][0] if param in self.query else default
        if specifier is None:
            return None
        if m := re.match(r'^\[(.*)]$', specifier):
            specifier = m.group(1)
        try:
            return pyproj.CRS(specifier)
        except pyproj.exceptions.CRSError as e:
            raise ValueError(e)



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Union

import xarray as xr

from xcube.core.mldataset import FsMultiLevelDataset
from xcube.core.mldataset import FsMultiLevelDatasetError
from xcube.core.mldataset import MultiLevelDataset
from xcube.core.subsampling import AGG_METHODS
# Note, we need the following reference to register the
# xarray property accessor
# noinspection PyUnresolvedReferences
from xcube.core.zarrstore import ZarrStoreHolder
from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonComplexSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonNullSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema
from .dataset import DatasetZarrFsDataAccessor
from ... import DataStoreError
from ...datatype import DATASET_TYPE
from ...datatype import DataType
from ...datatype import MULTI_LEVEL_DATASET_TYPE


class MultiLevelDatasetLevelsFsDataAccessor(DatasetZarrFsDataAccessor):
    """
    Opener/writer extension name "mldataset:levels:<protocol>".
    """

    @classmethod
    def get_data_type(cls) -> DataType:
        return MULTI_LEVEL_DATASET_TYPE

    @classmethod
    def get_format_id(cls) -> str:
        return 'levels'

    def open_data(self, data_id: str, **open_params) -> MultiLevelDataset:
        assert_instance(data_id, str, name='data_id')
        fs, fs_root, open_params = self.load_fs(open_params)
        try:
            return FsMultiLevelDataset(data_id,
                                       fs=fs,
                                       fs_root=fs_root,
                                       **open_params)
        except FsMultiLevelDatasetError as e:
            raise DataStoreError(f"{e}") from e

    def get_write_data_params_schema(self) -> JsonObjectSchema:
        schema = super().get_write_data_params_schema()  # creates deep copy
        # TODO: remove use_saved_levels, instead see #619
        schema.properties['use_saved_levels'] = JsonBooleanSchema(
            description='Whether to open an already saved level'
                        ' and downscale it then.'
                        ' May be used to avoid computation of'
                        ' entire Dask graphs at each level.',
            default=False,
        )
        schema.properties['base_dataset_id'] = JsonStringSchema(
            description='If given, avoids writing the base dataset'
                        ' at level 0. Instead a file "{data_id}/0.link"'
                        ' is created whose content is the given base dataset'
                        ' identifier.',
            nullable=True,
        )
        schema.properties['tile_size'] = JsonComplexSchema(
            one_of=[
                JsonIntegerSchema(minimum=1),
                JsonArraySchema(items=[
                    JsonIntegerSchema(minimum=1),
                    JsonIntegerSchema(minimum=1)
                ]),
                JsonNullSchema(),
            ],
            description='Tile size to be used for all levels of the'
                        ' written multi-level dataset.',
        )
        schema.properties['num_levels'] = JsonIntegerSchema(
            description='Maximum number of levels to be written.',
            minimum=1,
            nullable=True,
        )
        schema.properties['agg_methods'] = JsonComplexSchema(
            one_of=[
                JsonStringSchema(enum=AGG_METHODS),
                JsonObjectSchema(
                    additional_properties=JsonStringSchema(enum=AGG_METHODS)
                ),
                JsonNullSchema(),
            ],
            description='Aggregation method for the pyramid levels.'
                        ' If not explicitly set, "first" is used for integer '
                        ' variables and "mean" for floating point variables.'
                        ' If given as object, it is a mapping from variable '
                        ' name pattern to aggregation method. The pattern'
                        ' may include wildcard characters * and ?.'
        )
        return schema

    def write_data(self,
                   data: MultiLevelDataset,
                   data_id: str,
                   replace: bool = False,
                   **write_params) -> str:
        assert_instance(data, MultiLevelDataset, name='data')
        return self.write_data_generic(data,
                                       data_id,
                                       replace=replace,
                                       **write_params)

    def write_data_generic(self,
                           data: Union[xr.Dataset, MultiLevelDataset],
                           data_id: str,
                           replace: bool = False,
                           **write_params) -> str:
        assert_instance(data, (xr.Dataset, MultiLevelDataset), name='data')
        fs, fs_root, write_params = self.load_fs(write_params)
        base_dataset_id = write_params.pop("base_dataset_id", None)
        try:
            return FsMultiLevelDataset.write_dataset(
                data,
                data_id,
                fs=fs,
                fs_root=fs_root,
                replace=replace,
                base_dataset_path=base_dataset_id,
                **write_params
            )
        except FsMultiLevelDatasetError as e:
            raise DataStoreError(f"{e}") from e


class DatasetLevelsFsDataAccessor(MultiLevelDatasetLevelsFsDataAccessor):
    """
    Opener/writer extension name "dataset:levels:<protocol>".
    """

    @classmethod
    def get_data_type(cls) -> DataType:
        return DATASET_TYPE

    def open_data(self, data_id: str, **open_params) -> xr.Dataset:
        ml_dataset = super().open_data(data_id, **open_params)
        return ml_dataset.get_dataset(0)

    def write_data(self,
                   data: xr.Dataset,
                   data_id: str,
                   replace: bool = False,
                   **write_params) -> str:
        assert_instance(data, xr.Dataset, name='data')
        return self.write_data_generic(data,
                                       data_id,
                                       replace=replace,
                                       **write_params)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import copy
from abc import ABC, abstractmethod
from typing import Dict, Any, Tuple, Optional

import fsspec

from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonNumberSchema
from xcube.util.jsonschema import JsonObjectSchema
from ..accessor import DataOpener
from ..accessor import DataWriter
from ..datatype import DataType
from ..error import DataStoreError

COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES = dict(
    # passed to ``DirCache``, if the implementation supports
    # directory listing caching. Pass use_listings_cache=False
    # to disable such caching.
    use_listings_cache=JsonBooleanSchema(),
    listings_expiry_time=JsonNumberSchema(),
    max_paths=JsonIntegerSchema(),
    # If this is a cachable implementation, pass True here to force
    # creating a new instance even if a matching instance exists, and prevent
    # storing this instance.
    skip_instance_cache=JsonBooleanSchema(),
    asynchronous=JsonBooleanSchema(),
)

PROTOCOL_PARAM_NAME = 'protocol'
STORAGE_OPTIONS_PARAM_NAME = 'storage_options'
FS_PARAM_NAME = 'fs'
ROOT_PARAM_NAME = 'root'


class FsAccessor:
    """
    Base class for accessing some filesystem.
    """

    @classmethod
    def get_protocol(cls) -> str:
        """Get the filesystem protocol."""
        return 'abstract'

    @classmethod
    def get_storage_options_schema(cls) -> JsonObjectSchema:
        """Get the JSON schema of the filesystem parameters."""
        return JsonObjectSchema(
            properties=COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES,
            additional_properties=True,
        )

    @classmethod
    def load_fs(cls, params: Dict[str, Any]) \
            -> Tuple[fsspec.AbstractFileSystem,
                     Optional[str],
                     Dict[str, Any]]:
        """
        Load a filesystem instance from *params*.

        Pops useful parameters from a copy of *params* to get
        or instantiate the filesystem.
        Returns the filesystem and *params* reduced by
        the used parameters.

        :param params: Parameters passed to a filesystem
            data store, opener, or writer call.
        :return: A tuple comprising the filesystem, an optional root path,
            and the modified *params*.
        """
        params = dict(params)

        # Filesystem data-stores pass "fs" and "root" kwargs to
        # data opener and writer calls.
        fs = params.pop(FS_PARAM_NAME, None)
        root = params.pop(ROOT_PARAM_NAME, None)
        if fs is not None:
            assert_instance(fs, fsspec.AbstractFileSystem, name=FS_PARAM_NAME)
        if root is not None:
            assert_instance(root, str, name=ROOT_PARAM_NAME)
        if fs:
            return fs, root, params

        protocol = cls.get_protocol()
        if protocol == 'abstract':
            protocol = params.pop(PROTOCOL_PARAM_NAME, None)
            if protocol is None:
                raise DataStoreError(f"Cannot determine filesystem,"
                                     f" try using parameter"
                                     f" {PROTOCOL_PARAM_NAME!r}")

        storage_options = params.pop(STORAGE_OPTIONS_PARAM_NAME, None)
        if storage_options is not None:
            assert_instance(storage_options, dict,
                            name=STORAGE_OPTIONS_PARAM_NAME)
            storage_options = dict(storage_options)

        # Note, by default, filesystem data stores are writable and hence
        # SHALL NOT cache any directory listings!
        use_listings_cache = \
            bool(storage_options.pop('use_listings_cache', False)
                 if storage_options else False)

        try:
            return (
                fsspec.filesystem(protocol,
                                  use_listings_cache=use_listings_cache,
                                  **(storage_options or {})),
                root,
                params
            )
        except (ValueError, ImportError):
            raise DataStoreError(f"Cannot instantiate"
                                 f" filesystem {protocol!r}")

    @classmethod
    def add_storage_options_to_params_schema(
            cls, params_schema: JsonObjectSchema
    ) -> JsonObjectSchema:
        """
        Utility method to be used by subclasses to add the schema
        for the parameter "storage_options" to given *param_schema*.
        """
        params_schema = copy.deepcopy(params_schema)
        params_schema.properties[STORAGE_OPTIONS_PARAM_NAME] = \
            cls.get_storage_options_schema()
        return params_schema

    @classmethod
    def remove_storage_options_from_params_schema(
            cls, params_schema: JsonObjectSchema
    ) -> JsonObjectSchema:
        """
        Utility method to be used by subclasses to remove the schema
        for the parameter "storage_options" from given *param_schema*.
        """
        if STORAGE_OPTIONS_PARAM_NAME in params_schema.properties:
            params_schema = copy.deepcopy(params_schema)
            del params_schema.properties[STORAGE_OPTIONS_PARAM_NAME]
        return params_schema


class FsDataAccessor(DataOpener,
                     DataWriter,
                     FsAccessor,
                     ABC):
    """
    Abstract base class for data accessors that
    use an underlying filesystem.

    A ``FsDataAccessor`` is responsible for exactly one data type
    and one data format.

    Note, for all filesystem based accessors the parameter
    *data_id* is an absolute path that is build by the store
    according to ``f"{store.root}/{data_id}"``.
    The store's root may therefore only be useful to
    compute the original, relative store's *data_id*.
    """

    @classmethod
    @abstractmethod
    def get_data_type(cls) -> DataType:
        """
        Get the supported data type.
        """

    @classmethod
    @abstractmethod
    def get_format_id(cls) -> str:
        """
        Get the format identifier,
        for example "zarr" or "geojson".
        """

    def get_delete_data_params_schema(self, data_id: str = None) \
            -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                recursive=JsonBooleanSchema(),
                maxdepth=JsonIntegerSchema(),
                storage_options=self.get_storage_options_schema(),
            ),
            additional_properties=False,
        )

    def delete_data(self,
                    data_id: str,
                    **delete_params):
        fs, _, delete_params = self.load_fs(delete_params)
        fs.delete(data_id, **delete_params)



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Optional, Dict, Sequence

from xcube.util.undefined import UNDEFINED


class GeoJSON:
    PRIMITIVE_GEOMETRY_TYPES = {'Point', 'LineString', 'Polygon',
                                'MultiPoint', 'MultiLineString', 'MultiPolygon'}
    GEOMETRY_COLLECTION_TYPE = 'GeometryCollection'
    FEATURE_TYPE = 'Feature'
    FEATURE_COLLECTION_TYPE = 'FeatureCollection'

    @classmethod
    def is_point(cls, obj: Any) -> bool:
        return cls.get_type_name(obj) == 'Point'

    @classmethod
    def is_feature(cls, obj: Any) -> bool:
        return cls.get_type_name(obj) == cls.FEATURE_TYPE

    @classmethod
    def is_feature_collection(cls, obj: Any) -> bool:
        return cls.get_type_name(obj) == cls.FEATURE_COLLECTION_TYPE and cls._is_valid_sequence(obj, 'features')

    @classmethod
    def is_geometry(cls, obj: Any) -> bool:
        type_name = cls.get_type_name(obj)
        if type_name in cls.PRIMITIVE_GEOMETRY_TYPES:
            return cls._is_valid_sequence(obj, 'coordinates')
        if type_name == cls.GEOMETRY_COLLECTION_TYPE:
            return cls._is_valid_sequence(obj, 'geometries')
        return False

    @classmethod
    def is_geometry_collection(cls, obj: Any) -> bool:
        type_name = cls.get_type_name(obj)
        if type_name == cls.GEOMETRY_COLLECTION_TYPE:
            return cls._is_valid_sequence(obj, 'geometries')
        return False

    @classmethod
    def get_geometry_collection_geometries(cls, obj: Any) -> Optional[Sequence[Dict]]:
        type_name = cls.get_type_name(obj)
        if type_name == cls.GEOMETRY_COLLECTION_TYPE:
            geometries = cls._get_sequence(obj, 'geometries')
            if geometries == UNDEFINED:
                return None
            elif geometries is None:
                return []
            else:
                return geometries
        return None

    @classmethod
    def get_feature_collection_features(cls, obj: Any) -> Optional[Sequence[Dict]]:
        type_name = cls.get_type_name(obj)
        if type_name == cls.FEATURE_COLLECTION_TYPE:
            features = cls._get_sequence(obj, 'features')
            if features == UNDEFINED:
                return None
            elif features is None:
                return []
            else:
                return features
        return None

    @classmethod
    def get_feature_geometry(cls, obj: Any) -> Optional[Dict]:
        type_name = cls.get_type_name(obj)
        if type_name == cls.FEATURE_TYPE:
            if 'geometry' not in obj:
                return None
            geometry = obj['geometry']
            if cls.is_geometry(geometry):
                return geometry
            return None
        return None

    @classmethod
    def get_type_name(cls, obj: Any) -> Optional[str]:
        if not isinstance(obj, dict) or "type" not in obj:
            return None
        return obj["type"] or None

    @classmethod
    def _is_valid_sequence(cls, obj: Any, attr_name: str) -> bool:
        return cls._get_sequence(obj, attr_name) != UNDEFINED

    @classmethod
    def _get_sequence(cls, obj: Any, attr_name: str) -> Optional[Sequence[Any]]:
        if attr_name not in obj:
            return UNDEFINED
        sequence = obj[attr_name]
        if sequence is None:
            # GeoJSON sequence properties, 'coordinates', 'geometries', 'features', may be None according to spec
            return None
        try:
            iter(sequence)
            return sequence
        except TypeError:
            return UNDEFINED



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


from typing import Dict, List, Optional, Union, Sequence, Any, Set, Tuple

import numpy as np
import pandas as pd
import shapely.geometry
import xarray as xr

from xcube.constants import LOG
from xcube.core import timeseries
from xcube.core.ancvar import find_ancillary_var_names
from xcube.core.gridmapping import GridMapping
from xcube.server.api import ApiError
from xcube.util.geojson import GeoJSON
from xcube.util.perf import measure_time
from .context import TimeSeriesContext

TimeSeriesValue = Dict[str, Union[str, bool, int, float, None]]
TimeSeries = List[TimeSeriesValue]
TimeSeriesCollection = List[TimeSeries]
GeoJsonObj = Dict[str, Any]
GeoJsonFeature = GeoJsonObj
GeoJsonGeometry = GeoJsonObj


def get_time_series(
        ctx: TimeSeriesContext,
        ds_name: str,
        var_name: str,
        geo_json: GeoJsonObj,
        agg_methods: Union[str, Sequence[str]] = None,
        start_date: Optional[np.datetime64] = None,
        end_date: Optional[np.datetime64] = None,
        tolerance: Optional[float] = 1.0,
        max_valids: Optional[int] = None,
        incl_ancillary_vars: bool = False
) -> Union[TimeSeries, TimeSeriesCollection]:
    """Get the time-series for a given GeoJSON object *geo_json*.

    If *geo_json* is a single geometry or feature a list of
    time-series values is returned.
    If *geo_json* is a single geometry collection
    or feature collection a collection of lists of time-series values
    is returned so that geometry/feature collection and
    time-series collection elements are corresponding
    at same indices.

    A time series value always contains the key "time" whose
    value is an ISO date/time string. Other entries
    are values with varying keys depending on the geometry
    type and *agg_methods*. Their values may be
    either a bool, int, float or None.
    For point geometries the second key is "value".
    For non-point geometries that cover spatial areas, there
    will be values for all keys given by *agg_methods*.

    :param ctx: Service context object
    :param ds_name: The dataset identifier.
    :param var_name: The variable name.
    :param geo_json: The GeoJSON object that is a or has
        a geometry or collection of geometries.
    :param agg_methods: Spatial aggregation methods
        for geometries that cover a spatial area.
    :param start_date: An optional start date.
    :param end_date: An optional end date.
    :param tolerance: Time tolerance in seconds that expands
        the given time range. Defaults to one second.
    :param max_valids: Optional number of valid points.
        If it is None (default),
        also missing values are returned as NaN;
        if it is -1,
        only valid values are returned;
        if it is a positive integer,
        the most recent valid values are returned.
    :param incl_ancillary_vars: For point geometries,
        include values of ancillary variables, if any.
    :return: Time-series data structure.
    """
    if tolerance:
        timedelta = pd.Timedelta(tolerance, unit='seconds')
        if start_date is not None:
            start_date -= timedelta
        if end_date is not None:
            end_date += timedelta

    agg_methods = timeseries.normalize_agg_methods(
        agg_methods, exception_type=ApiError.BadRequest
    )

    ml_dataset = ctx.datasets_ctx.get_ml_dataset(ds_name)
    dataset = ctx.datasets_ctx.get_time_series_dataset(ds_name,
                                                       var_name=var_name)
    geo_json_geometries, is_collection = _to_geo_json_geometries(geo_json)
    geometries = _to_shapely_geometries(geo_json_geometries)

    with measure_time() as time_result:
        results = _get_time_series_for_geometries(
            dataset,
            var_name,
            geometries,
            start_date=start_date,
            end_date=end_date,
            agg_methods=agg_methods,
            grid_mapping=ml_dataset.grid_mapping,
            max_valids=max_valids,
            incl_ancillary_vars=incl_ancillary_vars
        )

    if ctx.datasets_ctx.trace_perf:
        LOG.info(f'get_time_series: dataset id {ds_name},'
                 f' variable {var_name}, '
                 f'{len(results)} x {len(results[0])} values,'
                 f' took {time_result.duration} seconds')

    return results[0] if not is_collection and len(results) == 1 else results


def _get_time_series_for_geometries(
        dataset: xr.Dataset,
        var_name: str,
        geometries: List[shapely.geometry.base.BaseGeometry],
        agg_methods: Set[str],
        grid_mapping: Optional[GridMapping] = None,
        start_date: Optional[np.datetime64] = None,
        end_date: Optional[np.datetime64] = None,
        max_valids: Optional[int] = None,
        incl_ancillary_vars: bool = False
) -> TimeSeriesCollection:
    time_series_collection = []
    for geometry in geometries:
        time_series = _get_time_series_for_geometry(
            dataset,
            var_name,
            geometry,
            agg_methods,
            grid_mapping=grid_mapping,
            start_date=start_date,
            end_date=end_date,
            max_valids=max_valids,
            incl_ancillary_vars=incl_ancillary_vars
        )
        time_series_collection.append(time_series)
    return time_series_collection


def _get_time_series_for_geometry(
        dataset: xr.Dataset,
        var_name: str,
        geometry: shapely.geometry.base.BaseGeometry,
        agg_methods: Set[str],
        grid_mapping: Optional[GridMapping] = None,
        start_date: Optional[np.datetime64] = None,
        end_date: Optional[np.datetime64] = None,
        max_valids: Optional[int] = None,
        incl_ancillary_vars: bool = False
) -> TimeSeries:
    if isinstance(geometry, shapely.geometry.Point):
        return _get_time_series_for_point(
            dataset,
            var_name,
            geometry,
            agg_methods,
            grid_mapping=grid_mapping,
            start_date=start_date,
            end_date=end_date,
            max_valids=max_valids,
            incl_ancillary_vars=incl_ancillary_vars
        )

    time_series_ds = timeseries.get_time_series(
        dataset,
        grid_mapping=grid_mapping,
        geometry=geometry,
        var_names=[var_name],
        agg_methods=agg_methods,
        start_date=start_date,
        end_date=end_date,
        cube_asserted=True
    )
    if time_series_ds is None:
        return []

    var_names = {agg_method: f'{var_name}_{agg_method}' for agg_method in
                 agg_methods}

    return collect_timeseries_result(time_series_ds,
                                     var_names,
                                     max_valids=max_valids)


def _get_time_series_for_point(
        dataset: xr.Dataset,
        var_name: str,
        point: shapely.geometry.Point,
        agg_methods: Set[str],
        grid_mapping: Optional[GridMapping] = None,
        start_date: Optional[np.datetime64] = None,
        end_date: Optional[np.datetime64] = None,
        max_valids: Optional[int] = None,
        incl_ancillary_vars: bool = False
) -> TimeSeries:
    var_key = None
    if timeseries.AGG_MEAN in agg_methods:
        var_key = timeseries.AGG_MEAN
    elif timeseries.AGG_MEDIAN in agg_methods:
        var_key = timeseries.AGG_MEDIAN
    elif timeseries.AGG_MIN in agg_methods \
            or timeseries.AGG_MAX in agg_methods:
        var_key = timeseries.AGG_MIN
    if not var_key:
        raise ApiError.BadRequest(
            'Aggregation methods must include one of'
            ' "mean", "median", "min", "max"'
        )

    roles_to_anc_var_names = dict()
    if incl_ancillary_vars:
        roles_to_anc_var_name_sets = find_ancillary_var_names(
            dataset,
            var_name,
            same_shape=True,
            same_dims=True
        )
        for role, roles_to_anc_var_name_sets \
                in roles_to_anc_var_name_sets.items():
            if role:
                roles_to_anc_var_names[
                    role] = roles_to_anc_var_name_sets.pop()

    var_names = [var_name] + list(set(roles_to_anc_var_names.values()))

    time_series_ds = timeseries.get_time_series(
        dataset,
        grid_mapping=grid_mapping,
        geometry=point,
        var_names=var_names,
        start_date=start_date,
        end_date=end_date,
        cube_asserted=True
    )
    if time_series_ds is None:
        return []

    key_to_var_names = {var_key: var_name}
    for role, anc_var_name in roles_to_anc_var_names.items():
        key_to_var_names[role] = anc_var_name

    return collect_timeseries_result(time_series_ds,
                                     key_to_var_names,
                                     max_valids=max_valids)


def collect_timeseries_result(time_series_ds: xr.Dataset,
                              key_to_var_names: Dict[str, str],
                              max_valids: int = None) -> TimeSeries:
    _check_max_valids(max_valids)

    var_values_map = dict()
    for key, var_name in key_to_var_names.items():
        values = time_series_ds[var_name].values
        if np.issubdtype(values.dtype, np.floating):
            num_type = float
        elif np.issubdtype(values.dtype, np.integer):
            num_type = int
        elif np.issubdtype(values.dtype, np.dtype(bool)):
            num_type = bool
        else:
            raise ValueError(f'cannot convert {values.dtype}'
                             f' into JSON-convertible value')
        var_values_map[key] = [(num_type(v) if f else None)
                               for f, v in zip(np.isfinite(values), values)]

    time_values = [
        t.isoformat() + 'Z'
        for t in pd.DatetimeIndex(time_series_ds.time.values).round('s')
    ]

    max_number_of_observations = time_series_ds.attrs.get(
        'max_number_of_observations', 1
    )
    num_times = len(time_values)
    time_series = []

    max_valids_is_positive = max_valids is not None and max_valids > 0
    if max_valids_is_positive:
        time_indexes = range(num_times - 1, -1, -1)
    else:
        time_indexes = range(num_times)

    for time_index in time_indexes:

        if len(time_series) == max_valids:
            break

        time_series_value = dict()
        all_null = True
        for key, var_values in var_values_map.items():
            var_value = var_values[time_index]
            if var_value is not None:
                all_null = False
            time_series_value[key] = var_value

        has_count = 'count' in time_series_value
        no_obs = all_null or (has_count and time_series_value['count'] == 0)
        if no_obs and max_valids is not None:
            continue

        time_series_value['time'] = time_values[time_index]
        if has_count:
            time_series_value['count_tot'] = max_number_of_observations

        time_series.append(time_series_value)

    if max_valids_is_positive:
        time_series = time_series[::-1]

    return time_series


def _to_shapely_geometries(geo_json_geometries: List[GeoJsonGeometry]) \
        -> List[shapely.geometry.base.BaseGeometry]:
    geometries = []
    for geo_json_geometry in geo_json_geometries:
        try:
            geometry = shapely.geometry.shape(geo_json_geometry)
        except (TypeError, ValueError) as e:
            raise ApiError.BadRequest(
                "Invalid GeoJSON geometry encountered"
            ) from e
        geometries.append(geometry)
    return geometries


def _to_geo_json_geometries(geo_json: GeoJsonObj) \
        -> Tuple[List[GeoJsonGeometry], bool]:
    is_collection = False
    if GeoJSON.is_feature(geo_json):
        geometry = _get_feature_geometry(geo_json)
        geometries = [geometry]
    elif GeoJSON.is_feature_collection(geo_json):
        is_collection = True
        features = GeoJSON.get_feature_collection_features(geo_json)
        geometries = [_get_feature_geometry(feature)
                      for feature in features] if features else []
    elif GeoJSON.is_geometry_collection(geo_json):
        is_collection = True
        geometries = GeoJSON.get_geometry_collection_geometries(geo_json)
    elif GeoJSON.is_geometry(geo_json):
        geometries = [geo_json]
    else:
        raise ApiError.BadRequest("GeoJSON object expected")
    return geometries, is_collection


def _get_feature_geometry(feature: GeoJsonFeature) -> GeoJsonGeometry:
    geometry = GeoJSON.get_feature_geometry(feature)
    if geometry is None or not GeoJSON.is_geometry(geometry):
        raise ApiError.BadRequest("GeoJSON feature without geometry")
    return geometry


def _get_float_value(values: Optional[np.ndarray], index: int) \
        -> Optional[float]:
    if values is None:
        return None
    value = float(values[index])
    return None if np.isnan(value) else value


def _check_max_valids(max_valids):
    if not (max_valids is None or max_valids == -1 or max_valids > 0):
        raise ApiError.BadRequest(
            f'max_valids must be either None, -1 or positive,'
            f' was {max_valids}'
        )



# The MIT License (MIT)
# Copyright (c) 2020-2024 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from xcube.core.store.fs.accessor import COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonComplexSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema

REF_STORE_SCHEMA = JsonObjectSchema(
    properties=dict(
        refs=JsonArraySchema(
            description=(
                "The list of references to use for this"
                " instance. Items can be a path / URL to a reference"
                " JSON file or dictionaries with keys `'ref_path'` and"
                " `'data_descriptor'`. Reference files (URLs or local paths)"
                " are used in conjunction with target_options and"
                " target_protocol to open and parse JSON at this location."
            ),
            items=JsonComplexSchema(
                one_of=[
                    JsonStringSchema(min_length=1),
                    JsonObjectSchema(
                        properties=dict(
                            ref_path=JsonStringSchema(min_length=1),
                            data_id=JsonStringSchema(min_length=1),
                            # We could detail data_descriptor a bit more. Not now.
                            data_descriptor=JsonObjectSchema(
                                additional_properties=True),
                        ),
                        required=["ref_path"]
                    ),
                ],
            ),
            min_items=1,
        ),
        target_protocol=JsonStringSchema(
            description=(
                "Used for loading the reference files."
                " If not given, protocol will be derived from the given path"
            )
        ),
        target_options=JsonObjectSchema(
            description=(
                "Extra filesystem options for loading the reference files."
            ),
            additional_properties=True
        ),
        remote_protocol=JsonStringSchema(
            description=(
                "The protocol of the filesystem on which the references"
                " will be evaluated. If not given, will be derived from"
                " the first URL in the references that has a protocol."
            )
        ),
        remote_options=JsonObjectSchema(
            description=(
                "Extra filesystem options for loading the referenced data."
            ),
            additional_properties=True
        ),
        max_gap=JsonIntegerSchema(
            description="See max_block."
        ),
        max_block=JsonIntegerSchema(
            description=(
                "For merging multiple concurrent requests to the same"
                " remote file. Neighboring byte ranges will only be"
                " merged when their inter-range gap is <= `max_gap`."
                " Default is 64KB. Set to 0 to only merge when it"
                " requires no extra bytes. Pass a negative number to"
                " disable merging, appropriate for local target files."
                " Neighboring byte ranges will only be merged when the"
                " size of the aggregated range is <= ``max_block``."
                " Default is 256MB."
            )
        ),
        cache_size=JsonIntegerSchema(
            description=(
                "Maximum size of LRU cache, where"
                " cache_size*record_size denotes the total number of"
                " references that can be loaded in memory at once."
                " Only used for lazily loaded references."
            )
        ),
        **COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES,
    ),
    additional_properties=True
)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .accessor import DataDeleter
from .accessor import DataOpener
from .accessor import DataTimeSliceUpdater
from .accessor import DataWriter
from .accessor import find_data_opener_extensions
from .accessor import find_data_writer_extensions
from .accessor import get_data_accessor_predicate
from .accessor import new_data_opener
from .accessor import new_data_writer
from .assertions import assert_valid_config
from .assertions import assert_valid_params
from .datatype import ANY_TYPE
from .datatype import DATASET_TYPE
from .datatype import DataType
from .datatype import DataTypeLike
from .datatype import GEO_DATA_FRAME_TYPE
from .datatype import MULTI_LEVEL_DATASET_TYPE
from .descriptor import DataDescriptor
from .descriptor import DatasetDescriptor
from .descriptor import GeoDataFrameDescriptor
from .descriptor import MultiLevelDatasetDescriptor
from .descriptor import VariableDescriptor
from .descriptor import new_data_descriptor
from .error import DataStoreError
from .fs.registry import new_fs_data_store
from .search import DataSearcher
from .search import DefaultSearchMixin
from .store import DataStore
from .store import MutableDataStore
from .store import find_data_store_extensions
from .store import get_data_store_class
from .store import get_data_store_params_schema
from .store import new_data_store
from .storepool import DataStoreConfig
from .storepool import DataStoreInstance
from .storepool import DataStorePool
from .storepool import DataStorePoolLike
from .storepool import get_data_store_instance



from typing import Dict, Union, Tuple, Any, Optional, \
    Mapping, Sequence, Hashable

import dask.array as da
import numpy as np
import pandas as pd
import xarray as xr

from xcube.core.verify import assert_cube

DEFAULT_INDEX_NAME_PATTERN = '{name}_index'
DEFAULT_REF_NAME_PATTERN = '{name}_ref'
INDEX_DIM_NAME = "idx"

POINT_INTERP_METHOD_NEAREST = "nearest"
POINT_INTERP_METHOD_LINEAR = "linear"
DEFAULT_INTERP_POINT_METHOD = POINT_INTERP_METHOD_NEAREST

SeriesLike = Union[
    np.ndarray,
    da.Array,
    xr.DataArray,
    pd.Series,
    Sequence[Union[int, float]]
]

PointsLike = Union[
    xr.Dataset,
    pd.DataFrame,
    Mapping[str, SeriesLike]
]


def get_cube_values_for_points(
        cube: xr.Dataset,
        points: PointsLike,
        var_names: Sequence[str] = None,
        include_coords: bool = False,
        include_bounds: bool = False,
        include_indexes: bool = False,
        index_name_pattern: str = DEFAULT_INDEX_NAME_PATTERN,
        include_refs: bool = False,
        ref_name_pattern: str = DEFAULT_REF_NAME_PATTERN,
        method: str = DEFAULT_INTERP_POINT_METHOD,
        cube_asserted: bool = False
) -> xr.Dataset:
    """
    Extract values from *cube* variables at given
    coordinates in *points*.

    Returns a new dataset with values of variables from *cube* selected
    by the coordinate columns
    provided in *points*. All variables will be 1-D and have the same order
    as the rows in points.

    :param cube: The cube dataset.
    :param points: Dictionary that maps dimension name to coordinate arrays.
    :param var_names: An optional list of names of data variables in *cube*
        whose values shall be extracted.
    :param include_coords: Whether to include the cube coordinates for each
        point in return value.
    :param include_bounds: Whether to include the cube coordinate boundaries
        (if any) for each point in return value.
    :param include_indexes: Whether to include computed indexes into the cube
        for each point in return value.
    :param index_name_pattern: A naming pattern for the computed index columns.
        Must include "{name}" which will be replaced by the index'
        dimension name.
    :param include_refs: Whether to include point (reference) values from
        *points* in return value.
    :param ref_name_pattern: A naming pattern for the computed point data
        columns. Must include "{name}" which will be replaced by the point's
        attribute name.
    :param method: "nearest" or "linear".
    :param cube_asserted: If False, *cube* will be verified, otherwise it
        is expected to be a valid cube.
    :return: A new data frame whose columns are values from *cube* variables
        at given *points*.
    """
    if not cube_asserted:
        assert_cube(cube)

    index_dtype = np.int64 \
        if method == POINT_INTERP_METHOD_NEAREST else np.float64

    point_indexes = get_cube_point_indexes(
        cube,
        points,
        index_name_pattern=index_name_pattern,
        index_dtype=index_dtype,
        cube_asserted=True
    )

    cube_values = get_cube_values_for_indexes(
        cube,
        point_indexes,
        include_coords,
        include_bounds,
        data_var_names=var_names,
        index_name_pattern=index_name_pattern,
        method=method,
        cube_asserted=True
    )

    if include_indexes:
        cube_values.update(point_indexes)

    if include_refs:
        point_refs = xr.Dataset({
            ref_name_pattern.format(name=name):
                xr.DataArray(points[name], dims=[INDEX_DIM_NAME])
            for name in points.keys()
        })

        cube_values.update(point_refs)

    # The (coordinate) variable named INDEX_DIM_NAME should not exist,
    # otherwise cube_values.to_dataframe() will use an empty string
    # to name that series "" in the output DataFrame instance.
    # This seems to be an issue of xarray 0.18.2+.
    if INDEX_DIM_NAME in cube_values:
        cube_values = cube_values.drop_vars(INDEX_DIM_NAME)

    return cube_values


def get_cube_values_for_indexes(
        cube: xr.Dataset,
        indexes: Union[xr.Dataset, pd.DataFrame, Mapping[str, Any]],
        include_coords: bool = False,
        include_bounds: bool = False,
        data_var_names: Sequence[str] = None,
        index_name_pattern: str = DEFAULT_INDEX_NAME_PATTERN,
        method: str = DEFAULT_INTERP_POINT_METHOD,
        cube_asserted: bool = False
) -> xr.Dataset:
    """
    Get values from the *cube* at given *indexes*.

    :param cube: A cube dataset.
    :param indexes: A mapping from column names to index and fraction arrays
        for all cube dimensions.
    :param include_coords: Whether to include the cube coordinates for each
        point in return value.
    :param include_bounds: Whether to include the cube coordinate boundaries
        (if any) for each point in return value.
    :param data_var_names: An optional list of names of data variables in
        *cube* whose values shall be extracted.
    :param index_name_pattern: A naming pattern for the computed
        indexes columns. Must include "{name}" which will be replaced by
        the dimension name.
    :param method: "nearest" or "linear".
    :param cube_asserted: If False, *cube* will be verified, otherwise it
        is expected to be a valid cube.
    :return: A new data frame whose columns are values from *cube* variables
        at given *indexes*.
    """
    if not cube_asserted:
        assert_cube(cube)

    if method not in {POINT_INTERP_METHOD_NEAREST, POINT_INTERP_METHOD_LINEAR}:
        raise ValueError(f"invalid method {method!r}")
    if method != POINT_INTERP_METHOD_NEAREST:
        raise NotImplementedError(f"method {method!r} not yet implemented")

    all_data_var_names = tuple(cube.data_vars.keys())
    if len(all_data_var_names) == 0:
        raise ValueError("cube is empty")

    if data_var_names is not None:
        if len(data_var_names) == 0:
            return _empty_dataset_from_points(indexes)
        for var_name in data_var_names:
            if var_name not in cube.data_vars:
                raise ValueError(f"variable {var_name!r} not found in cube")
    else:
        data_var_names = all_data_var_names

    dim_names = cube[data_var_names[0]].dims
    num_dims = len(dim_names)
    index_names = [index_name_pattern.format(name=dim_name)
                   for dim_name in dim_names]
    indexes, num_points = _normalize_series(indexes,
                                            index_names,
                                            force_dataset=True,
                                            param_name="indexes")
    if num_points == 0:
        return _empty_dataset_from_points(indexes)
    cube = xr.Dataset({var_name: cube[var_name]
                       for var_name in data_var_names}, coords=cube.coords)

    new_bounds_vars = {}
    bounds_var_names = _get_coord_bounds_var_names(cube)
    drop_coords = None
    if bounds_var_names:
        if include_bounds:
            # Flatten any coordinate bounds variables
            for var_name, bnds_var_name in bounds_var_names.items():
                bnds_var = cube[bnds_var_name]
                new_bounds_vars[f"{var_name}_lower"] = bnds_var[:, 0]
                new_bounds_vars[f"{var_name}_upper"] = bnds_var[:, 1]
            cube = cube.assign_coords(**new_bounds_vars)
        cube = cube.drop_vars(bounds_var_names.values())
        if not include_coords:
            drop_coords = set(cube.coords).difference(new_bounds_vars.keys())
    else:
        if not include_coords:
            drop_coords = set(cube.coords)

    # Generate a validation condition so we can
    # filter out invalid rows (where any index == -1)
    is_valid_point = None
    for index_name in index_names:
        col = indexes[index_name]
        condition = col >= 0 \
            if np.issubdtype(col.dtype, np.integer) else np.isnan(col)
        if is_valid_point is None:
            is_valid_point = condition
        else:
            is_valid_point = np.logical_and(is_valid_point, condition)

    num_valid_points = np.count_nonzero(is_valid_point)
    if num_valid_points == num_points:
        # All indexes valid
        cube_selector = {dim_names[i]: indexes[index_names[i]]
                         for i in range(num_dims)}
        cube_values = cube.isel(cube_selector)
    elif num_valid_points == 0:
        # All indexes are invalid
        new_bounds_vars = {}
        for var_name in cube.variables:
            new_bounds_vars[var_name] = _empty_points_var(cube[var_name],
                                                          num_points)
        cube_values = xr.Dataset(new_bounds_vars)
    else:
        # Some invalid indexes
        idx = np.arange(num_points)
        good_idx = idx[is_valid_point.values]
        idx_dim_name = indexes[index_names[0]].dims[0]
        good_indexes = indexes.isel({idx_dim_name: good_idx})

        cube_selector = {dim_names[i]: good_indexes[index_names[i]]
                         for i in range(num_dims)}
        cube_values = cube.isel(cube_selector)

        new_bounds_vars = {}
        for var_name in cube.variables:
            var = cube_values[var_name]
            new_var = _empty_points_var(var, num_points)
            new_var[good_idx] = var
            new_bounds_vars[var_name] = new_var

        cube_values = xr.Dataset(new_bounds_vars)

    if drop_coords:
        cube_values = cube_values.drop_vars(drop_coords)

    return cube_values


def get_cube_point_indexes(
        cube: xr.Dataset,
        points: PointsLike,
        dim_name_mapping: Mapping[str, str] = None,
        index_name_pattern: str = DEFAULT_INDEX_NAME_PATTERN,
        index_dtype=np.float64,
        cube_asserted: bool = False
) -> xr.Dataset:
    """
    Get indexes of given point coordinates *points* into the given *dataset*.

    :param cube: The cube dataset.
    :param points: A mapping from column names to column data arrays, which
        must all have the same length.
    :param dim_name_mapping: A mapping from dimension names in *cube* to
        column names in *points*.
    :param index_name_pattern: A naming pattern for the computed
        indexes columns. Must include "{name}" which will be replaced by
        the dimension name.
    :param index_dtype: Numpy data type for the indexes. If it is a
        floating point type (default),
        then *indexes* will contain fractions, which may be
        used for interpolation.
        For out-of-range coordinates in *points*, indexes will be -1
        if *index_dtype* is an integer type, and NaN,
        if *index_dtype* is a floating point types.
    :param cube_asserted: If False, *cube* will be verified, otherwise
        it is expected to be a valid cube.
    :return: A dataset containing the index columns.
    """
    if not cube_asserted:
        assert_cube(cube)

    dim_name_mapping = dim_name_mapping if dim_name_mapping is not None else {}
    dim_names = _get_cube_data_var_dims(cube)
    col_names = [dim_name_mapping.get(str(dim_name), dim_name)
                 for dim_name in dim_names]

    points, _ = _normalize_series(points,
                                  col_names,
                                  force_dataset=False,
                                  param_name="points")

    indexes = []
    for dim_name, col_name in zip(dim_names, col_names):
        col = points[col_name]
        coord_indexes = get_dataset_indexes(cube,
                                            str(dim_name),
                                            col,
                                            index_dtype=index_dtype)
        indexes.append((index_name_pattern.format(name=dim_name),
                        xr.DataArray(coord_indexes, dims=[INDEX_DIM_NAME])))

    return xr.Dataset(dict(indexes))


def get_dataset_indexes(
        dataset: xr.Dataset,
        coord_var_name: str,
        coord_values: SeriesLike,
        index_dtype=np.float64
) -> Union[xr.DataArray, np.ndarray]:
    """
    Compute the indexes and their fractions into a coordinate variable
    *coord_var_name* of a *dataset* for the given coordinate values
    *coord_values*.

    The coordinate variable's labels must be monotonic increasing or
    decreasing, otherwise the result will be nonsense.

    For any value in *coord_values* that is out of the bounds of the
    coordinate variable's values, the index depends on the value of
    *index_dtype*. If *index_dtype* is an integer type, invalid indexes
    are encoded as -1 while for floating point types, NaN will be used.

    Returns a tuple of indexes as int64 array and fractions as
    float64 array.

    :param dataset: A cube dataset.
    :param coord_var_name: Name of a coordinate variable.
    :param coord_values: Array-like coordinate values.
    :param index_dtype: Numpy data type for the indexes.
        If it is floating point type (default), then *indexes*
        contain fractions, which may be used for interpolation.
        If *dtype* is an integer type out-of-range coordinates
        are indicated by index -1, and NaN if it is is a
        floating point type.
    :return: The indexes and their fractions as a tuple of
        numpy int64 and float64 arrays.
    """
    coord_var = dataset[coord_var_name]
    n1 = coord_var.size
    n2 = n1 + 1

    coord_bounds_var = _get_bounds_var(dataset, coord_var_name)
    if coord_bounds_var is not None:
        coord_bounds = coord_bounds_var.values
        if np.issubdtype(coord_bounds.dtype, np.datetime64):
            coord_bounds = coord_bounds.astype(np.uint64)
        is_inverse = (coord_bounds[0, 1] - coord_bounds[0, 0]) < 0
        if is_inverse:
            coord_bounds = coord_bounds[::-1, ::-1]
        coords = np.zeros(n2, dtype=coord_bounds.dtype)
        coords[0:-1] = coord_bounds[:, 0]
        coords[-1] = coord_bounds[-1, 1]
    elif coord_var.size > 1:
        center_coords = coord_var.values
        if np.issubdtype(center_coords.dtype, np.datetime64):
            center_coords = center_coords.astype(np.uint64)
        is_inverse = (center_coords[-1] - center_coords[0]) < 0
        if is_inverse:
            center_coords = center_coords[::-1]
        deltas = np.zeros(n2, dtype=center_coords.dtype)
        deltas[0:-2] = np.diff(center_coords)
        deltas[-2] = deltas[-3]
        deltas[-1] = deltas[-3]
        coords = np.zeros(n2, dtype=center_coords.dtype)
        coords[0:-1] = center_coords
        coords[-1] = coords[-2] + deltas[-1]
        if np.issubdtype(deltas.dtype, np.integer):
            coords -= deltas // 2
        else:
            coords -= 0.5 * deltas
    else:
        raise ValueError(f"cannot determine cell boundaries for"
                         f" coordinate variable {coord_var_name!r}"
                         f" of size {coord_var.size}")

    if np.issubdtype(coord_values.dtype, np.datetime64):
        try:
            coord_values = coord_values.astype(np.uint64)
        except TypeError:
            # Fixes https://github.com/dcs4cop/xcube/issues/95
            coord_values = coord_values.values.astype(np.uint64)

    indexes = np.linspace(0.0, n1, n2, dtype=np.float64)
    interp_indexes = np.interp(coord_values, coords, indexes,
                               left=-1, right=-1)
    if is_inverse:
        i = interp_indexes >= 0
        interp_indexes[i] = n1 - interp_indexes[i]
    interp_indexes[interp_indexes >= n1] = n1 - 1e-9
    if np.issubdtype(index_dtype, np.integer):
        interp_indexes = interp_indexes.astype(index_dtype)
    else:
        interp_indexes[interp_indexes < 0] = np.nan

    return interp_indexes


def _empty_points_var(var: xr.DataArray, num_points: int):
    fill_value = 0 if np.issubdtype(var.dtype, np.integer) else np.nan
    return xr.DataArray(np.full(num_points, fill_value, dtype=var.dtype),
                        dims=[INDEX_DIM_NAME], attrs=var.attrs)


def _normalize_series(
        points: PointsLike,
        col_names: Sequence[str],
        force_dataset: bool = False,
        param_name: str = "points"
) -> Tuple[PointsLike, int]:
    if not isinstance(points, (xr.Dataset, pd.DataFrame)):
        new_points = {}
        for col_name in col_names:
            if col_name in points:
                col = points[col_name]
                if not isinstance(col, xr.DataArray):
                    col = xr.DataArray(col)
                new_points[col_name] = col
        points = new_points

    num_points = None
    for col_name in col_names:
        if col_name not in points:
            raise ValueError(f"column {col_name!r} not found in {param_name}")
        col = points[col_name]
        if len(col.shape) != 1:
            raise ValueError(f"column {col_name!r} in {param_name} must be"
                             f" one-dimensional, but has shape {col.shape!r}")
        if num_points is None:
            num_points = len(col)
        elif num_points != len(col):
            raise ValueError(f"column sizes in {param_name} must be all the"
                             f" same, but found {len(points[col_names[0]])}"
                             f" for column {col_names[0]!r} and {num_points}"
                             f" for column {col_name!r}")

    if num_points is None:
        raise ValueError(f"{param_name} has no valid columns")

    if force_dataset:
        if isinstance(points, pd.DataFrame):
            points = xr.Dataset.from_dataframe(points)
        elif not isinstance(points, xr.DataArray):
            points = xr.Dataset(points)

    return points, num_points


def _get_coord_bounds_var_names(dataset: xr.Dataset) -> Dict[Hashable, Any]:
    bnds_var_names = {}
    for var_name in dataset.coords:
        var = dataset[var_name]
        bnds_var_name = var.attrs.get("bounds", f"{var_name}_bnds")
        if bnds_var_name not in bnds_var_names and bnds_var_name in dataset:
            bnds_var = dataset[bnds_var_name]
            if bnds_var.shape == (var.shape[0], 2):
                bnds_var_names[var_name] = bnds_var_name
    return bnds_var_names


def _get_cube_data_var_dims(cube: xr.Dataset) -> Tuple[Hashable, ...]:
    for var in cube.data_vars.values():
        return var.dims
    raise ValueError("cube dataset is empty")


def _get_bounds_var(dataset: xr.Dataset,
                    var_name: str) -> Optional[xr.DataArray]:
    var = dataset[var_name]
    if len(var.shape) == 1:
        bounds_var_name = var.attrs.get("bounds", f"{var_name}_bnds")
        if bounds_var_name in dataset:
            bounds_var = dataset[bounds_var_name]
            if bounds_var.dtype == var.dtype \
                    and bounds_var.shape == (var.size, 2):
                return bounds_var
    return None


def _empty_dataset_from_points(data: Any) -> xr.Dataset:
    return xr.Dataset(
        coords=data.coords if hasattr(data, "coords") else None
    )



# The MIT License (MIT)
# Copyright (c) 2021/2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import asyncio
import json
import socket
import threading
import time
import unittest
from abc import ABC
from contextlib import closing
from typing import Dict, Optional, Any, Union, Tuple

import urllib3
from tornado.platform.asyncio import AnyThreadEventLoopPolicy
from urllib3.exceptions import MaxRetryError

from xcube.server.server import Server
from xcube.server.webservers.tornado import TornadoFramework
from xcube.util.extension import ExtensionRegistry


# taken from https://stackoverflow.com/a/45690594
def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('localhost', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


class ServerTestCase(unittest.TestCase, ABC):

    def setUp(self) -> None:
        self.port = find_free_port()
        config = self.get_config()
        config.update({
            'port': self.port,
            'address': 'localhost'
        })
        self.add_config(config)
        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
        er = self.get_extension_registry()
        self.add_extension(er)
        self.server = Server(framework=TornadoFramework(), config=config,
                             extension_registry=er)
        tornado = threading.Thread(target=self.server.start)
        tornado.daemon = True
        tornado.start()

        self.http = urllib3.PoolManager()

        # Taking care that server is fully started before tests make requests.
        # Fixing https://github.com/dcs4cop/xcube/issues/899
        while True:
            try:
                self.fetch('/')
                break
            except (MaxRetryError, TimeoutError):
                time.sleep(0.1)

    def tearDown(self) -> None:
        self.server.stop()

    # noinspection PyMethodMayBeStatic
    def get_extension_registry(self) -> ExtensionRegistry:
        return ExtensionRegistry()

    def add_extension(self, er: ExtensionRegistry) -> None:
        pass

    # noinspection PyMethodMayBeStatic
    def get_config(self) -> Dict[str, Any]:
        return {}

    def add_config(self, config: Dict) -> None:
        pass

    def fetch(self,
              path: str,
              method: str = "GET",
              headers: Optional[Dict[str, str]] = None,
              body: Optional[Any] = None) \
            -> urllib3.response.HTTPResponse:
        """Fetches the response for given request.

        :param path: Route request path
        :param method: HTTP request method, default to "GET"
        :param headers: HTTP request headers
        :param body: HTTP request body
        :return: A response object
        """
        if path.startswith("/"):
            path = path[1:]
        url = f'http://localhost:{self.port}/{path}'
        if isinstance(body, dict):
            body = bytes(json.dumps(body), 'utf-8')
        return self.http.request(method, url, headers=headers, body=body)

    def fetch_json(self,
                   path: str,
                   method: str = "GET",
                   headers: Optional[Dict[str, str]] = None,
                   body: Optional[Any] = None) \
            -> Tuple[
                Union[type(None), bool, int, float, str, list, dict], int]:
        """Fetches the response's JSON value for given request.

        :param path: Route request path
        :param method: HTTP request method, default to "GET"
        :param headers: HTTP request headers
        :param body: HTTP request body
        :return: A json object parsed from response data
        """
        response = self.fetch(path, method=method, headers=headers, body=body)
        return json.loads(response.data), response.status

    def assertResponseOK(
            self,
            response: urllib3.response.HTTPResponse
    ):
        """ Assert that response has status code 200.

        :param response: The response from a ``self.fetch()`` call
        """
        self.assertResponse(response, expected_status=200)

    def assertResourceNotFoundResponse(
            self,
            response: urllib3.response.HTTPResponse,
            expected_message: str = None
    ):
        """ Assert that response
        has the given status code 404, and
        has the given error message *expected_message*, if given.

        :param response: The response from a ``self.fetch()`` call
        :param expected_message: The expected error message or part of it.
        """
        self.assertResponse(response,
                            expected_status=404,
                            expected_message=expected_message)

    def assertBadRequestResponse(
            self,
            response: urllib3.response.HTTPResponse,
            expected_message: str = None
    ):
        """ Assert that response
        has the given status code 400, and
        has the given error message *expected_message*, if given.

        :param response: The response from a ``self.fetch()`` call
        :param expected_message: The expected error message or part of it.
        """
        self.assertResponse(response,
                            expected_status=400,
                            expected_message=expected_message)

    def assertResponse(
            self,
            response: urllib3.response.HTTPResponse,
            expected_status: Optional[int] = None,
            expected_message: Optional[str] = None
    ):
        """ Assert that response
        has the given status code *expected_status*, if given, and
        has the given error message *expected_message*, if given.

        :param response: The response from a ``self.fetch()`` call
        :param expected_status: The expected status code.
        :param expected_message: The expected error message or part of it.
        """
        message = self.parse_error_message(response)
        if expected_status is not None:
            self.assertEqual(expected_status, response.status,
                             msg=message or response.reason)
        if expected_message is not None:
            self.assertIsNotNone(message,
                                 msg="error message not available")
            self.assertIsInstance(message, str,
                                  msg="error message must be a string,"
                                      f" was {message!r}")
            self.assertIn(expected_message, message)

    @classmethod
    def parse_error_message(
            cls,
            response: urllib3.response.HTTPResponse
    ) -> Optional[str]:
        """Parse an error message from the given response object."""
        # noinspection PyBroadException
        try:
            result = json.loads(response.data)
            return result['error']['message']
        except BaseException:
            return None


# TODO (forman): Remove before final PR
ServerTest = ServerTestCase



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import collections.abc
import inspect
import itertools
import json
import math
import threading
import warnings
from typing import Iterator, Dict, Tuple, Any, Callable, \
    Optional, List, Sequence
from typing import Union

import numcodecs.abc
import numpy as np
import xarray as xr
import zarr.storage

from xcube.util.assertions import assert_instance, assert_true

GetData = Callable[[Tuple[int]],
                   Union[bytes, np.ndarray]]

OnClose = Callable[[Dict[str, Any]], None]


class GenericArray(Dict[str, any]):
    """
    Represent a generic array in the ``GenericZarrStore`` as
    dictionary of properties.

    Although all properties of this class are optional,
    some of them are mandatory when added to the ``GenericZarrStore``.

    When added to the store using ``GenericZarrStore.add_array()``,
    the array *name* and *dims* must always be present.
    Other mandatory properties depend on
    the *data* and *get_data* properties, which are mutually exclusive:

    * *get_data* is called for a requested data chunk of an array.
      It must return a bytes object or a numpy nd-array and is passed
      the chunk index, the chunk shape, and this array info dictionary.
      *get_data* requires the following properties to be present too:
      *name*, *dims*, *dtype*, *shape*.
      *chunks* is optional and defaults to *shape*.
    * *data* must be a bytes object or a numpy nd-array.
      *data* requires the following properties to be present too:
      *name*, *dims*. *chunks* must be same as *shape*.

    The function *get_data* receives only keyword-arguments which
    comprises the ones passed by *get_data_params*, if any, and
    two special ones which may occur in the signature of *get_data*:

    * The keyword argument *chunk_info*, if given, provides a dictionary
      that holds information about the current chunk:
      - ``index: tuple[int, ...]`` - the chunk's index
      - ``shape: tuple[int, ...]`` - the chunk's shape
      - ``slices: tuple[slice, ...]`` - the chunk's array slices

    * The keyword argument *array_info*, if given, provides a dictionary
      that holds information about the overall array. It contains
      all array properties passed to the constructor of ``GenericArray``
      plus
      - ``ndim: int`` - number of dimensions
      - ``num_chunks: tuple[int, ...]`` - number of chunks in every dimension

    ``GenericZarrStore`` will convert a Numpy array returned
    by *get_data* or given by *data* into a bytes object.
    It will also be compressed, if a *compressor* is given.
    It is important that the array chunks always See also
    https://zarr.readthedocs.io/en/stable/spec/v2.html#chunks

    Note that if the value of a named keyword argument is None,
    it will not be stored.

    :param array: Optional array info dictionary
    :param name: Optional array name
    :param data: Optional array data.
        Mutually exclusive with *get_data*.
        Must be a bytes object or a numpy array.
    :param get_data: Optional array data chunk getter.
        Mutually exclusive with *data*.
        Called for a requested data chunk of an array.
        Must return a bytes object or a numpy array.
    :param get_data_params: Optional keyword-arguments passed
        to *get_data*.
    :param dtype: Optional array data type.
        Either a string using syntax of the Zarr spec or a ``numpy.dtype``.
        For string encoded data types, see
        https://zarr.readthedocs.io/en/stable/spec/v2.html#data-type-encoding
    :param dims: Optional sequence of dimension names.
    :param shape: Optional sequence of shape sizes for each dimension.
    :param chunks: Optional sequence of chunk sizes for each dimension.
    :param fill_value: Optional fill value, see
        https://zarr.readthedocs.io/en/stable/spec/v2.html#fill-value-encoding
    :param compressor: Optional compressor.
        If given, it must be an instance of ``numcodecs.abc.Codec``.
    :param filters: Optional sequence of filters, see
        https://zarr.readthedocs.io/en/stable/spec/v2.html#filters.
    :param order: Optional array endian ordering.
        If given, must be "C" or "F". Defaults to "C".
    :param attrs: Optional array attributes.
        If given, must be JSON-serializable.
    :param on_close: Optional array close handler.
        Called if the store is closed.
    :param chunk_encoding: Optional encoding type of the chunk
        data returned for the array. Can be "bytes" (the default)
        or "ndarray" for array chunks that are numpy.ndarray instances.
    :param kwargs: Other keyword arguments passed directly to the
        dictionary constructor.
    """

    def __init__(self,
                 array: Optional[Dict[str, any]] = None,
                 name: Optional[str] = None,
                 get_data: Optional[GetData] = None,
                 get_data_params: Optional[Dict[str, Any]] = None,
                 data: Optional[np.ndarray] = None,
                 dtype: Optional[Union[str, np.dtype]] = None,
                 dims: Optional[Union[str, Sequence[str]]] = None,
                 shape: Optional[Sequence[int]] = None,
                 chunks: Optional[Sequence[int]] = None,
                 fill_value: Optional[Union[bool, int, float, str]] = None,
                 compressor: Optional[numcodecs.abc.Codec] = None,
                 filters: Optional[Sequence[numcodecs.abc.Codec]] = None,
                 order: Optional[str] = None,
                 attrs: Optional[Dict[str, Any]] = None,
                 on_close: Optional[OnClose] = None,
                 chunk_encoding: Optional[str] = None,
                 **kwargs):
        array = dict(array) if array is not None else dict()
        array.update({
            k: v
            for k, v in dict(
                name=name,
                dtype=dtype,
                dims=dims,
                shape=shape,
                chunks=chunks,
                fill_value=fill_value,
                compressor=compressor,
                filters=filters,
                order=order,
                attrs=attrs,
                data=data,
                get_data=get_data,
                get_data_params=get_data_params,
                on_close=on_close,
                chunk_encoding=chunk_encoding
            ).items()
            if v is not None
        })
        super().__init__(array, **kwargs)

    def finalize(self) -> "GenericArray":
        """Normalize and validate array properties and return a valid
        array info dictionary to be stored in the `GenericZarrStore`.
        """
        name = self.get("name")
        if not name:
            raise ValueError("missing array name")

        data = self.get("data")
        get_data = self.get("get_data")
        if data is None and get_data is None:
            raise ValueError(f"array {name!r}:"
                             f" either data or get_data must be defined")
        if get_data is not None:
            if data is not None:
                raise ValueError(f"array {name!r}:"
                                 f" data and get_data cannot"
                                 f" be defined together")
            if not callable(get_data):
                raise TypeError(f"array {name!r}:"
                                f" get_data must be a callable")
            sig = inspect.signature(get_data)
            get_data_info = {
                "has_array_info": "array_info" in sig.parameters,
                "has_chunk_info": "chunk_info" in sig.parameters,
            }
            get_data_params = dict(self.get("get_data_params") or {})
        else:
            get_data_info = None
            get_data_params = None

        dims = self.get("dims")
        dims = [dims] if isinstance(dims, str) else dims
        if dims is None:
            raise ValueError(f"array {name!r}: missing dims")

        ndim = len(dims)

        if isinstance(data, np.ndarray):
            # forman: maybe warn if dtype or shape is given,
            #  but does not match data.dtype and data.shape
            dtype = str(data.dtype.str)
            shape = data.shape
            chunks = data.shape
        else:
            dtype = self.get("dtype")
            shape = self.get("shape")
            chunks = self.get("chunks", shape)

        if not dtype:
            raise ValueError(f"array {name!r}: missing dtype")
        elif isinstance(dtype, np.dtype):
            dtype = dtype.str

        if shape is None:
            raise ValueError(f"array {name!r}: missing shape")
        if len(shape) != ndim:
            raise ValueError(f"array {name!r}:"
                             f" dims and shape must have same length")
        if len(chunks) != ndim:
            raise ValueError(f"array {name!r}:"
                             f" dims and chunks must have same length")

        num_chunks = tuple(map(lambda x: math.ceil(x[0] / x[1]),
                               zip(shape, chunks)))

        filters = self.get("filters")
        if filters:
            filters = list(filters)
            for f in filters:
                if not isinstance(f, numcodecs.abc.Codec):
                    raise TypeError(f"array {name!r}:"
                                    f" filter items must be an"
                                    f" instance of numcodecs.abc.Codec")
        else:
            filters = None

        compressor = self.get("compressor")
        if compressor is not None:
            if not isinstance(compressor, numcodecs.abc.Codec):
                raise TypeError(f"array {name!r}:"
                                f" compressor must be an"
                                f" instance of numcodecs.abc.Codec")

        fill_value = self.get("fill_value")
        if isinstance(fill_value, np.ndarray):
            fill_value = fill_value.item()
        allowed_fill_value_types = (type(None), bool, int, float, str)
        if not isinstance(fill_value, allowed_fill_value_types):
            raise TypeError(
                f"array {name!r}:"
                f" fill_value type must be one of"
                f" {tuple(t.__name__ for t in allowed_fill_value_types)},"
                f" was {type(fill_value).__name__}"
            )

        order = self.get("order") or "C"
        allowed_orders = ("C", "F")
        if order not in allowed_orders:
            raise ValueError(
                f"array {name!r}:"
                f" order must be one of {allowed_orders},"
                f" was {order!r}"
            )

        chunk_encoding = self.get("chunk_encoding") or "bytes"
        allowed_chunk_encodings = ("bytes", "ndarray")
        if chunk_encoding not in allowed_chunk_encodings:
            raise ValueError(
                f"array {name!r}:"
                f" chunk_encoding must be one of {allowed_chunk_encodings},"
                f" was {chunk_encoding!r}"
            )

        attrs = self.get("attrs")
        if attrs is not None:
            if not isinstance(attrs, dict):
                raise TypeError(
                    f"array {name!r}:"
                    f" attrs must be dict, was {type(attrs).__name__}"
                )

        # Note: passing the properties as dictionary
        # will prevent removing them if their value is None,
        # see GenericArray constructor.
        return GenericArray({
            "name": name,
            "dtype": dtype,
            "dims": tuple(dims),
            "shape": tuple(shape),
            "chunks": tuple(chunks),
            "fill_value": fill_value,
            "filters": filters,
            "compressor": compressor,
            "order": order,
            "attrs": attrs,
            "data": data,
            "get_data": get_data,
            "get_data_params": get_data_params,
            "on_close": self.get("on_close"),
            "chunk_encoding": chunk_encoding,
            # Computed properties
            "ndim": len(dims),
            "num_chunks": num_chunks,
            "get_data_info": get_data_info,
        })


GenericArrayLike = Union[GenericArray, Dict[str, Any]]


class GenericZarrStore(zarr.storage.Store):
    """A Zarr store that maintains generic arrays in a flat, top-level
    hierarchy. The root of the store is a Zarr group
    conforming to the Zarr spec v2.

    It is designed to serve as a Zarr store for xarray datasets
    that compute their data arrays dynamically.

    See class ``GenericArray`` for specifying the arrays' properties.

    The array data of this store's arrays are either retrieved from
    static (numpy) arrays or from a callable that provides the
    array's data chunks as bytes or numpy arrays.

    :param arrays: Arrays to be added.
        Typically, these will be instances of ``GenericArray``.
    :param attrs: Optional attributes of the top-level group.
        If given, it must be JSON serializable.
    :param array_defaults: Optional array defaults for
        array properties not passed to ``add_array``.
        Typically, this will be an instance of ``GenericArray``.
    """

    # Shortcut for GenericArray
    Array = GenericArray

    def __init__(
            self,
            *arrays: GenericArrayLike,
            attrs: Optional[Dict[str, Any]] = None,
            array_defaults: Optional[GenericArrayLike] = None
    ):
        self._attrs = dict(attrs) if attrs is not None else {}
        self._array_defaults = array_defaults
        self._dim_sizes: Dict[str, int] = {}
        self._arrays: Dict[str, GenericArray] = {}
        for array in arrays:
            self.add_array(array)

    def add_array(self,
                  array: Optional[GenericArrayLike] = None,
                  **array_kwargs) -> None:
        """
        Add a new array to this store.

        :param array: Optional array properties.
            Typically, this will be an instance of ``GenericArray``.
        :param array_kwargs: Keyword arguments form
            for the properties of ``GenericArray``.
        """
        effective_array = GenericArray(self._array_defaults or {})
        if array:
            effective_array.update(array)
        if array_kwargs:
            effective_array.update(array_kwargs)
        effective_array = effective_array.finalize()

        name = effective_array["name"]
        if name in self._arrays:
            raise ValueError(f"array {name!r} is already defined")

        dims = effective_array["dims"]
        shape = effective_array["shape"]
        for dim_name, dim_size in zip(dims, shape):
            old_dim_size = self._dim_sizes.get(dim_name)
            if old_dim_size is None:
                self._dim_sizes[name] = dim_size
            elif old_dim_size != dim_size:
                # Dimensions must have same lengths for all arrays
                # in this store
                raise ValueError(f"array {name!r}"
                                 f" defines dimension {dim_name!r}"
                                 f" with size {dim_size},"
                                 f" but existing size is {old_dim_size}")

        self._arrays[name] = effective_array

    ##########################################################################
    # Zarr Store implementation
    ##########################################################################

    def is_writeable(self) -> bool:
        """Return False, because arrays in this store are generative."""
        return False

    def listdir(self, path: str = "") -> List[str]:
        """List a store path.
        :param path: The path.
        :return: List of sorted directory entries.
        """
        if path == "":
            return sorted([
                ".zmetadata",
                ".zgroup",
                ".zattrs",
                *self._arrays.keys()
            ])
        elif "/" not in path:
            return sorted(self._get_array_keys(path))
        raise ValueError(f"{path} is not a directory")

    def rmdir(self, path: str = "") -> None:
        """The general form removes store paths.
        This implementation can remove entire arrays only.
        :param path: The array's name.
        """
        if path not in self._arrays:
            raise ValueError(f"{path}: can only remove existing arrays")
        array = self._arrays.pop(path)
        dims = array["dims"]
        for i, dim_name in enumerate(dims):
            dim_used = False
            for array_name, array in self._arrays.items():
                if dim_name in array["dims"]:
                    dim_used = True
                    break
            if not dim_used:
                del self._dim_sizes[dim_name]

    def rename(self, src_path: str, dst_path: str) -> None:
        """The general form renames store paths.
        This implementation can rename arrays only.

        :param src_path: Source array name.
        :param dst_path: Target array name.
        """
        array = self._arrays.get(src_path)
        if array is None:
            raise ValueError(f"can only rename arrays, but {src_path!r}"
                             f" is not an array")
        if dst_path in self._arrays:
            raise ValueError(f"cannot rename array {src_path!r} into"
                             f" {dst_path!r} because it already exists")
        if "/" in dst_path:
            raise ValueError(f"cannot rename array {src_path!r}"
                             f" into {dst_path!r}")
        array["name"] = dst_path
        self._arrays[dst_path] = array
        del self._arrays[src_path]

    def close(self) -> None:
        """Calls the "on_close" handlers, if any, of arrays."""
        for array in self._arrays.values():
            on_close = array.get("on_close")
            if on_close is not None:
                on_close(array)

    # Note, getsize is not implemented by intention as it requires
    # actual computation of arrays.
    #
    # def getsize(self, key: str) -> int:
    #    pass

    ##########################################################################
    # MutableMapping implementation
    ##########################################################################

    def __iter__(self) -> Iterator[str]:
        """Get an iterator of all keys in this store."""
        yield ".zmetadata"
        yield ".zgroup"
        yield ".zattrs"
        for array_name in self._arrays.keys():
            yield from self._get_array_keys(array_name)

    def __len__(self) -> int:
        return sum(1 for _ in iter(self))

    def __contains__(self, key: str) -> bool:
        if key in (".zmetadata", ".zgroup", ".zattrs"):
            return True
        try:
            array_name, value_id = self._parse_array_key(key)
        except KeyError:
            return False
        if value_id in (".zarray", ".zattrs"):
            return True
        try:
            self._get_array_chunk_index(array_name, value_id)
            return True
        except KeyError:
            return False

    def __getitem__(self, key: str) -> Union[bytes, np.ndarray]:
        item = self._get_item(key)
        if isinstance(item, dict):
            return dict_to_bytes(item)
        elif isinstance(item, str):
            return str_to_bytes(item)
        return item

    def __setitem__(self, key: str, value: bytes) -> None:
        class_name = self.__module__ + '.' + self.__class__.__name__
        raise TypeError(f'{class_name} is read-only')

    def __delitem__(self, key: str) -> None:
        self.rmdir(key)

    ########################################################################
    # Utilities
    ##########################################################################

    @classmethod
    def from_dataset(cls,
                     dataset: xr.Dataset,
                     array_defaults: Optional[GenericArrayLike] = None) \
            -> "GenericZarrStore":
        """Create a Zarr store for given *dataset*.
        to the *dataset*'s attributes.
        The following *array_defaults* properties can be provided
        (other properties are prescribed by the *dataset*):

        * ``fill_value``- defaults to None
        * ``compressor``- defaults to None
        * ``filters``- defaults to None
        * ``order``- defaults to "C"
        * ``chunk_encoding`` - defaults to "bytes"

        :param dataset: The dataset
        :param array_defaults: Array default values.
        :return: A new Zarr store instance.
        """

        def _get_dataset_data(ds=None,
                              chunk_info=None,
                              array_info=None) -> np.ndarray:
            array_name = array_info["name"]
            chunk_slices = chunk_info["slices"]
            return ds[array_name][chunk_slices].values

        arrays = []
        for var_name, var in dataset.variables.items():
            arrays.append(GenericArray(
                name=str(var_name),
                dtype=np.dtype(var.dtype).str,
                dims=[str(dim) for dim in var.dims],
                shape=var.shape,
                chunks=[(max(*c) if len(c) > 1 else c[0])
                        for c in var.chunks] if var.chunks else var.shape,
                attrs={str(k): v for k, v in var.attrs.items()},
                get_data=_get_dataset_data,
                get_data_params=dict(ds=dataset),
            ))

        attrs = {str(k): v for k, v in dataset.attrs.items()}
        return GenericZarrStore(*arrays,
                                attrs=attrs,
                                array_defaults=array_defaults)

    ########################################################################
    # Helpers
    ##########################################################################

    def _get_item(self, key: str) -> Union[dict, str, bytes]:
        if key == ".zmetadata":
            return self._get_metadata_item()
        if key == ".zgroup":
            return self._get_group_item()
        if key == ".zattrs":
            return self._get_attrs_item()

        array_name, value_id = self._parse_array_key(key)
        array = self._arrays[array_name]

        if value_id == '.zarray':
            return self._get_array_spec_item(array)
        if value_id == '.zattrs':
            return self._get_array_attrs_item(array)

        chunk_index = self._get_array_chunk_index(array_name, value_id)
        return self._get_array_data_item(array, chunk_index)

    def _get_metadata_item(self):
        metadata = {
            ".zgroup": self._get_item(".zgroup"),
            ".zattrs": self._get_item(".zattrs"),
        }
        for array_name in self._arrays.keys():
            key = array_name + "/.zarray"
            metadata[key] = self._get_item(key)
            key = array_name + "/.zattrs"
            metadata[key] = self._get_item(key)
        return {
            "zarr_consolidated_format": 1,
            "metadata": metadata
        }

    # noinspection PyMethodMayBeStatic
    def _get_group_item(self):
        return {
            "zarr_format": 2
        }

    def _get_attrs_item(self):
        return self._attrs or {}

    # noinspection PyMethodMayBeStatic
    def _get_array_spec_item(self, array: GenericArray):

        # JSON-encode fill_value
        fill_value = array["fill_value"]
        if isinstance(fill_value, float):
            if math.isnan(fill_value):
                fill_value = "NaN"
            elif math.isinf(fill_value):
                if fill_value < 0:
                    fill_value = "-Infinity"
                else:
                    fill_value = "Infinity"

        # JSON-encode compressor
        compressor = array["compressor"]
        if compressor is not None:
            compressor = compressor.get_config()

        # JSON-encode filters
        filters = array["filters"]
        if filters is not None:
            filters = list(f.get_config() for f in filters)

        return {
            "zarr_format": 2,
            "dtype": array["dtype"],
            "shape": list(array["shape"]),
            "chunks": list(array["chunks"]),
            "fill_value": fill_value,
            "compressor": compressor,
            "filters": filters,
            "order": array["order"],
        }

    # noinspection PyMethodMayBeStatic
    def _get_array_attrs_item(self, array: GenericArray):
        dims = array["dims"]
        attrs = array["attrs"]
        return {
            "_ARRAY_DIMENSIONS": dims,
            **(attrs or {})
        }

    # noinspection PyMethodMayBeStatic
    def _get_array_data_item(self,
                             array: Dict[str, Any],
                             chunk_index: Tuple[int]) \
            -> Union[bytes, np.ndarray]:
        # Note, here array is expected to be "finalized",
        # that is, validated and normalized

        shape = array["shape"]
        chunks = array["chunks"]
        chunk_shape = None

        data = array["data"]
        if data is None:
            get_data = array["get_data"]
            assert callable(get_data)  # Has been ensured before
            get_data_params = array["get_data_params"]
            get_data_kwargs = dict(get_data_params)
            get_data_info = array["get_data_info"]
            if get_data_info["has_chunk_info"]:
                chunk_shape = get_chunk_shape(shape, chunks, chunk_index)
                array_slices = get_array_slices(shape, chunks, chunk_index)
                get_data_kwargs["chunk_info"] = {
                    "index": chunk_index,
                    "shape": chunk_shape,
                    "slices": array_slices,
                }
            if get_data_info["has_array_info"]:
                get_data_kwargs["array_info"] = dict(array)

            data = get_data(**get_data_kwargs)

        chunk_encoding = array["chunk_encoding"]
        if isinstance(data, np.ndarray):
            # As of Zarr 2.0, all chunks of an array
            # must have the same shape (= chunks)
            if data.shape != chunks:
                # This commonly happens if array shape sizes
                # are not integer multiple of chunk shape sizes.
                if chunk_shape is None:
                    # Compute expected chunk shape.
                    chunk_shape = get_chunk_shape(shape, chunks, chunk_index)
                # We will only pad the data if the data shape
                # corresponds to the expected chunk's shape.
                if data.shape == chunk_shape:
                    padding = get_chunk_padding(shape, chunks, chunk_index)
                    fill_value = array["fill_value"]
                    data = np.pad(data, padding,
                                  mode="constant",
                                  constant_values=fill_value or 0)
                else:
                    key = format_chunk_key(array["name"], chunk_index)
                    raise ValueError(f"{key}:"
                                     f" data chunk at {chunk_index}"
                                     f" must have shape {chunk_shape},"
                                     f" but was {data.shape}")
            if chunk_encoding == "bytes":
                # Convert to bytes, filter and compress
                data = ndarray_to_bytes(data,
                                        order=array["order"],
                                        filters=array["filters"],
                                        compressor=array["compressor"])

        # Sanity check
        if (chunk_encoding == "bytes"
            and not isinstance(data, bytes)) \
                or (chunk_encoding == "ndarray"
                    and not isinstance(data, np.ndarray)):
            key = format_chunk_key(array["name"],
                                   chunk_index)
            expected_type = "numpy.ndarray" if chunk_encoding == "ndarray" \
                else "bytes"
            raise TypeError(f"{key}:"
                            f" data must be encoded as {expected_type},"
                            f" but was {type(data).__name__}")

        return data

    def _parse_array_key(self, key: str) -> Tuple[str, str]:
        array_name_and_value_id = key.rsplit('/', maxsplit=1)
        if len(array_name_and_value_id) != 2:
            raise KeyError(key)
        array_name, value_id = array_name_and_value_id
        if array_name not in self._arrays:
            raise KeyError(key)
        return array_name, value_id

    def _get_array_chunk_index(self,
                               array_name: str,
                               index_id: str) -> Tuple[int]:
        try:
            chunk_index = tuple(map(int, index_id.split('.')))
        except (ValueError, TypeError):
            raise KeyError(f"{array_name}/{index_id}")
        array = self._arrays[array_name]
        shape = array["shape"]
        if len(chunk_index) != len(shape):
            raise KeyError(f"{array_name}/{index_id}")
        num_chunks = array["num_chunks"]
        for i, n in zip(chunk_index, num_chunks):
            if not (0 <= i < n):
                raise KeyError(f"{array_name}/{index_id}")
        return chunk_index

    def _get_array_keys(self, array_name: str) -> Iterator[str]:
        yield array_name + "/.zarray"
        yield array_name + "/.zattrs"
        array = self._arrays[array_name]
        num_chunks = array["num_chunks"]
        yield from get_chunk_keys(array_name, num_chunks)


def get_array_slices(shape: Tuple[int, ...],
                     chunks: Tuple[int, ...],
                     chunk_index: Tuple[int, ...]) -> Tuple[slice, ...]:
    return tuple(
        slice(i * c,
              i * c + (c if (i + 1) * c <= s else s % c))
        for s, c, i in zip(shape, chunks, chunk_index)
    )


def get_chunk_shape(shape: Tuple[int, ...],
                    chunks: Tuple[int, ...],
                    chunk_index: Tuple[int, ...]) -> Tuple[int, ...]:
    return tuple(
        c if (i + 1) * c <= s else s % c
        for s, c, i in zip(shape, chunks, chunk_index)
    )


def get_chunk_padding(shape: Tuple[int, ...],
                      chunks: Tuple[int, ...],
                      chunk_index: Tuple[int, ...]):
    return tuple(
        (0, 0 if (i + 1) * c <= s else c - s % c)
        for s, c, i in zip(shape, chunks, chunk_index)
    )


def get_chunk_indexes(num_chunks: Tuple[int, ...]) \
        -> Iterator[Tuple[int, ...]]:
    if not num_chunks:
        yield 0,
    else:
        yield from itertools.product(*tuple(map(range, map(int, num_chunks))))


def get_chunk_keys(array_name: str,
                   num_chunks: Tuple[int, ...]) -> Iterator[str]:
    for chunk_index in get_chunk_indexes(num_chunks):
        yield format_chunk_key(array_name, chunk_index)


def format_chunk_key(array_name: str,
                     chunk_index: Tuple[int, ...]) -> str:
    chunk_id = '.'.join(map(str, chunk_index))
    return f"{array_name}/{chunk_id}"


def dict_to_bytes(d: Dict) -> bytes:
    return str_to_bytes(json.dumps(d, indent=2))


def str_to_bytes(s: str) -> bytes:
    return bytes(s, encoding='utf-8')


def ndarray_to_bytes(
        data: np.ndarray,
        order: Optional[str] = None,
        filters: Optional[Sequence[Any]] = None,
        compressor: Optional[numcodecs.abc.Codec] = None
) -> bytes:
    data = data.tobytes(order=order or "C")
    if filters:
        for f in filters:
            data = f.encode(data)
    if compressor is not None:
        data = compressor.encode(data)
    return data




# The MIT License (MIT)
# Copyright (c) 2020 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import inspect
import warnings
from typing import Tuple, Sequence, Dict, Any, Callable, Union, AbstractSet

import numpy as np
import xarray as xr

from xcube.core.schema import CubeSchema
from xcube.core.chunkstore import ChunkStore
from xcube.core.verify import assert_cube

CubeFuncOutput = Union[xr.DataArray, np.ndarray, Sequence[Union[xr.DataArray, np.ndarray]]]
CubeFunc = Callable[..., CubeFuncOutput]

_PREDEFINED_KEYWORDS = ['input_params', 'dim_coords', 'dim_ranges']


# TODO: support vectorize = all cubes have same variables and cube_func receives variables as vectors (with extra dim)

def compute_cube(cube_func: CubeFunc,
                 *input_cubes: xr.Dataset,
                 input_cube_schema: CubeSchema = None,
                 input_var_names: Sequence[str] = None,
                 input_params: Dict[str, Any] = None,
                 output_var_name: str = 'output',
                 output_var_dtype: Any = np.float64,
                 output_var_attrs: Dict[str, Any] = None,
                 vectorize: bool = None,
                 cube_asserted: bool = False) -> xr.Dataset:
    """
    Compute a new output data cube with a single variable named *output_var_name*
    from variables named *input_var_names* contained in zero, one, or more
    input data cubes in *input_cubes* using a cube factory function *cube_func*.

    For a more detailed description of the function usage,
    please refer to :func:compute_dataset.

    :param cube_func: The cube factory function.
    :param input_cubes: An optional sequence of input cube datasets, must be provided if *input_cube_schema* is not.
    :param input_cube_schema: An optional input cube schema, 
    must be provided if *input_cubes* is not. 
    Will be ignored if *input_cubes* is provided.
    :param input_var_names: A sequence of variable names
    :param input_params: Optional dictionary with processing parameters passed to *cube_func*.
    :param output_var_name: Optional name of the output variable, defaults to ``'output'``.
    :param output_var_dtype: Optional numpy datatype of the output variable, defaults to ``'float32'``.
    :param output_var_attrs: Optional metadata attributes for the output variable.
    :param vectorize: Whether all *input_cubes* have the same variables which are concatenated and passed as vectors
        to *cube_func*. Not implemented yet.
    :param cube_asserted: If False, *cube* will be verified, otherwise it is expected to be a valid cube.
    :return: A new dataset that contains the computed output variable.
    """
    return compute_dataset(cube_func,
                           *input_cubes,
                           input_cube_schema=input_cube_schema,
                           input_var_names=input_var_names,
                           input_params=input_params,
                           output_var_name=output_var_name,
                           output_var_dtype=output_var_dtype,
                           output_var_attrs=output_var_attrs,
                           vectorize=vectorize,
                           cube_asserted=cube_asserted)


def compute_dataset(cube_func: CubeFunc,
                    *input_cubes: xr.Dataset,
                    input_cube_schema: CubeSchema = None,
                    input_var_names: Sequence[str] = None,
                    input_params: Dict[str, Any] = None,
                    output_var_name: str = 'output',
                    output_var_dims: AbstractSet[str] = None,
                    output_var_dtype: Any = np.float64,
                    output_var_attrs: Dict[str, Any] = None,
                    vectorize: bool = None,
                    cube_asserted: bool = False) -> xr.Dataset:
    """
    Compute a new output dataset with a single variable named *output_var_name*
    from variables named *input_var_names* contained in zero, one, or more
    input data cubes in *input_cubes* using a cube factory function *cube_func*.

    *cube_func* is called concurrently for each of the chunks of the input variables.
    It is expected to return a chunk block whith is type ``np.ndarray``.

    If *input_cubes* is not empty, *cube_func* receives variables as specified by *input_var_names*.
    If *input_cubes* is empty, *input_var_names* must be empty too, and *input_cube_schema*
    must be given, so that a new cube can be created.

    The full signature of *cube_func* is:::

        def cube_func(*input_vars: np.ndarray,
                      input_params: Dict[str, Any] = None,
                      dim_coords: Dict[str, np.ndarray] = None,
                      dim_ranges: Dict[str, Tuple[int, int]] = None) -> np.ndarray:
            pass

    The arguments are:

    * ``input_vars``: the variables according to the given *input_var_names*;
    * ``input_params``: is this call's *input_params*, a mapping from parameter name to value;
    * ``dim_coords``: a mapping from dimension names to the current chunk's coordinate arrays;
    * ``dim_ranges``: a mapping from dimension names to the current chunk's index ranges.

    Only the ``input_vars`` argument is mandatory. The keyword arguments
    ``input_params``, ``input_params``, ``input_params`` do need to be present at all.

    *output_var_dims* my be given in the case, where ...
    TODO: describe new output_var_dims...

    :param cube_func: The cube factory function.
    :param input_cubes: An optional sequence of input cube datasets, must be provided if *input_cube_schema* is not.
    :param input_cube_schema: An optional input cube schema, must be provided if *input_cubes* is not.
    :param input_var_names: A sequence of variable names
    :param input_params: Optional dictionary with processing parameters passed to *cube_func*.
    :param output_var_name: Optional name of the output variable, defaults to ``'output'``.
    :param output_var_dims: Optional set of names of the output dimensions,
        used in the case *cube_func* reduces dimensions.
    :param output_var_dtype: Optional numpy datatype of the output variable, defaults to ``'float32'``.
    :param output_var_attrs: Optional metadata attributes for the output variable.
    :param vectorize: Whether all *input_cubes* have the same variables which are concatenated and passed as vectors
        to *cube_func*. Not implemented yet.
    :param cube_asserted: If False, *cube* will be verified, otherwise it is expected to be a valid cube.
    :return: A new dataset that contains the computed output variable.
    """
    if vectorize is not None:
        # TODO: support vectorize = all cubes have same variables and cube_func
        #       receives variables as vectors (with extra dim)
        raise NotImplementedError('vectorize is not supported yet')

    if not cube_asserted:
        for cube in input_cubes:
            assert_cube(cube)

    # Check compatibility of inputs
    if input_cubes:
        input_cube_schema = CubeSchema.new(input_cubes[0])
        for cube in input_cubes:
            if not cube_asserted:
                assert_cube(cube)
            if cube != input_cubes[0]:
                # noinspection PyUnusedLocal
                other_schema = CubeSchema.new(cube)
                # TODO (forman): broadcast all cubes to same shape, rechunk to same chunks
    elif input_cube_schema is None:
        raise ValueError('input_cube_schema must be given')

    output_var_name = output_var_name or 'output'

    # Collect named input variables, raise if not found
    input_var_names = input_var_names or []
    input_vars = []
    for var_name in input_var_names:
        input_var = None
        for cube in input_cubes:
            if var_name in cube.data_vars:
                input_var = cube[var_name]
                break
        if input_var is None:
            raise ValueError(f'variable {var_name!r} not found in any of cubes')
        input_vars.append(input_var)

    # Find out, if cube_func uses any of _PREDEFINED_KEYWORDS
    has_input_params, has_dim_coords, has_dim_ranges = _inspect_cube_func(cube_func, input_var_names)

    def cube_func_wrapper(index_chunk, *input_var_chunks):
        nonlocal input_cube_schema, input_var_names, input_params, input_vars
        nonlocal has_input_params, has_dim_coords, has_dim_ranges

        # Note, xarray.apply_ufunc does a test call with empty input arrays,
        # so index_chunk.size == 0 is a valid case
        empty_call = index_chunk.size == 0

        # TODO: when output_var_dims is given, index_chunk must be reordered
        #   as core dimensions are moved to the and of index_chunk and input_var_chunks
        if not empty_call:
            index_chunk = index_chunk.ravel()

        if index_chunk.size < 2 * input_cube_schema.ndim:
            if not empty_call:
                warnings.warn(f"unexpected index_chunk of size {index_chunk.size} received!")
                return None

        dim_ranges = None
        if has_dim_ranges or has_dim_coords:
            dim_ranges = {}
            for i in range(input_cube_schema.ndim):
                dim_name = input_cube_schema.dims[i]
                if not empty_call:
                    start = int(index_chunk[2 * i + 0])
                    end = int(index_chunk[2 * i + 1])
                    dim_ranges[dim_name] = start, end
                else:
                    dim_ranges[dim_name] = ()

        dim_coords = None
        if has_dim_coords:
            dim_coords = {}
            for coord_var_name, coord_var in input_cube_schema.coords.items():
                coord_slices = [slice(None)] * coord_var.ndim
                for i in range(input_cube_schema.ndim):
                    dim_name = input_cube_schema.dims[i]
                    if dim_name in coord_var.dims:
                        j = coord_var.dims.index(dim_name)
                        coord_slices[j] = slice(*dim_ranges[dim_name])
                dim_coords[coord_var_name] = coord_var[tuple(coord_slices)].values

        kwargs = {}
        if has_input_params:
            kwargs['input_params'] = input_params
        if has_dim_ranges:
            kwargs['dim_ranges'] = dim_ranges
        if has_dim_coords:
            kwargs['dim_coords'] = dim_coords

        return cube_func(*input_var_chunks, **kwargs)

    index_var = _gen_index_var(input_cube_schema)

    all_input_vars = [index_var] + input_vars

    input_core_dims = None
    if output_var_dims:
        input_core_dims = []
        has_warned = False
        for i in range(len(all_input_vars)):
            input_var = all_input_vars[i]
            var_core_dims = [dim for dim in input_var.dims if dim not in output_var_dims]
            must_rechunk = False
            if var_core_dims and input_var.chunks:
                for var_core_dim in var_core_dims:
                    dim_index = input_var.dims.index(var_core_dim)
                    dim_chunk_size = input_var.chunks[dim_index][0]
                    dim_shape_size = input_var.shape[dim_index]
                    if dim_chunk_size != dim_shape_size:
                        must_rechunk = True
                        break
            if must_rechunk:
                if not has_warned:
                    warnings.warn(f'Input variables must not be chunked in dimension(s): {", ".join(var_core_dims)}.\n'
                                  f'Rechunking applies, which may drastically decrease runtime performance '
                                  f'and increase memory usage.')
                    has_warned = True
                all_input_vars[i] = input_var.chunk({var_core_dim: -1 for var_core_dim in var_core_dims})
            input_core_dims.append(var_core_dims)

    output_var = xr.apply_ufunc(cube_func_wrapper,
                                *all_input_vars,
                                dask='parallelized',
                                input_core_dims=input_core_dims,
                                output_dtypes=[output_var_dtype])
    if output_var_attrs:
        output_var.attrs.update(output_var_attrs)
    return xr.Dataset({output_var_name: output_var}, coords=input_cube_schema.coords)


def _inspect_cube_func(cube_func: CubeFunc, input_var_names: Sequence[str] = None):
    args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations \
        = inspect.getfullargspec(cube_func)
    cube_func_name = '?'
    if hasattr(cube_func, '__name__'):
        cube_func_name = cube_func.__name__
    true_args = [arg not in _PREDEFINED_KEYWORDS for arg in args]
    if False in true_args and any(true_args[true_args.index(False):]):
        raise ValueError(f'invalid cube_func {cube_func_name!r}: '
                         f'any argument must occur before any of {", ".join(_PREDEFINED_KEYWORDS)}, '
                         f'but got {", ".join(args)}')
    if not all(true_args) and varargs:
        raise ValueError(f'invalid cube_func {cube_func_name!r}: '
                         f'any argument must occur before any of {", ".join(_PREDEFINED_KEYWORDS)}, '
                         f'but got {", ".join(args)} before *{varargs}')
    num_input_vars = len(input_var_names) if input_var_names else 0
    num_args = sum(true_args)
    if varargs is None and num_input_vars != num_args:
        raise ValueError(f'invalid cube_func {cube_func_name!r}: '
                         f'expected {num_input_vars} arguments, '
                         f'but got {", ".join(args)}')
    has_input_params = 'input_params' in args or 'input_params' in kwonlyargs
    has_dim_coords = 'dim_coords' in args or 'dim_coords' in kwonlyargs
    has_dim_ranges = 'dim_ranges' in args or 'dim_ranges' in kwonlyargs
    return has_input_params, has_dim_coords, has_dim_ranges


def _gen_index_var(cube_schema: CubeSchema):
    dims = cube_schema.dims
    shape = cube_schema.shape
    chunks = cube_schema.chunks

    # noinspection PyUnusedLocal
    def get_chunk(cube_store: ChunkStore, name: str, index: Tuple[int, ...]) -> bytes:
        data = np.zeros(cube_store.chunks, dtype=np.uint64)
        data_view = data.ravel()
        if data_view.base is not data:
            raise ValueError('view expected')
        if data_view.size < cube_store.ndim * 2:
            raise ValueError('size too small')
        for i in range(cube_store.ndim):
            j1 = cube_store.chunks[i] * index[i]
            j2 = j1 + cube_store.chunks[i]
            data_view[2 * i] = j1
            data_view[2 * i + 1] = j2
        return data.tobytes()

    store = ChunkStore(dims, shape, chunks)
    store.add_lazy_array('__index_var__', '<u8', get_chunk=get_chunk)

    dataset = xr.open_zarr(store)
    index_var = dataset.__index_var__
    index_var = index_var.assign_coords(**cube_schema.coords)
    return index_var



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import importlib.resources
import os
import pathlib
from typing import Union, Optional

from xcube.constants import LOG
from xcube.server.api import ApiError
from xcube.server.api import ApiHandler
from .api import api
from .context import ViewerContext

ENV_VAR_XCUBE_VIEWER_PATH = "XCUBE_VIEWER_PATH"


@api.route('/viewer/config/{*path}')
class ViewerConfigHandler(ApiHandler[ViewerContext]):
    @api.operation(operationId='getViewerConfigurationItem',
                   summary='Get a configuration item for the xcube viewer.')
    def get(self, path: Optional[str]):
        config_items = self.ctx.config_items
        if config_items is None:
            raise ApiError.NotFound(f'xcube viewer has'
                                    f' not been been configured')
        try:
            data = config_items[path]
        except KeyError:
            raise ApiError.NotFound(f'The item {path!r} was not found'
                                    f' in viewer configuration')
        content_type = self.get_content_type(path)
        self.response.set_header("Content-Length", str(len(data)))
        if content_type is not None:
            self.response.set_header("Content-Type", content_type)
        self.response.write(data)

    @staticmethod
    def get_content_type(path: str) -> Optional[str]:
        filename_ext = path.split('/')[-1].split('.')[-1]
        if filename_ext in ("json", "xml"):
            return f"application/{filename_ext}; charset=UTF-8"
        elif filename_ext in ("csv", "html", "css", "js"):
            return f"text/{filename_ext}; charset=UTF-8"
        elif filename_ext in ("geojson",):
            return "application/geo+json; charset=UTF-8"
        elif filename_ext in ("gif", "png"):
            return f"image/{filename_ext}"
        elif filename_ext in ("jpeg", "jpg"):
            return "image/jpeg"
        return None


_responses = {
    200: {
        "description": "OK",
        "content": {
            "text/html": {
                "schema": {
                    "type": "string"
                },
            },
        }
    }
}

_viewer_module = "xcube.webapi.viewer"
_data_dir = "data"
_default_filename = "index.html"


@api.static_route('/viewer',
                  default_filename=_default_filename,
                  summary="Brings up the xcube Viewer webpage",
                  responses=_responses)
def get_local_viewer_path() -> Union[None, str, pathlib.Path]:
    local_viewer_path = os.environ.get(ENV_VAR_XCUBE_VIEWER_PATH)
    if local_viewer_path:
        return local_viewer_path
    try:
        with importlib.resources.files(_viewer_module) as path:
            local_viewer_path = path / _data_dir
            if (local_viewer_path / _default_filename).is_file():
                return local_viewer_path
    except ImportError:
        pass
    LOG.warning(f"Cannot find {_data_dir}/{_default_filename}"
                f" in {_viewer_module}, consider setting environment variable"
                f" {ENV_VAR_XCUBE_VIEWER_PATH}",
                exc_info=True)
    return None



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from xcube.server.api import ApiHandler
from .api import api
from .context import TilesContext
from .controllers import compute_ml_dataset_tile
from ..datasets import PATH_PARAM_DATASET_ID
from ..datasets import PATH_PARAM_VAR_NAME
from ..datasets import QUERY_PARAM_CBAR
from ..datasets import QUERY_PARAM_CRS
from ..datasets import QUERY_PARAM_VMAX
from ..datasets import QUERY_PARAM_VMIN

PATH_PARAM_X = {
    "name": "x",
    "in": "path",
    "description": "The tile grid's x-coordinate",
    "schema": {
        "type": "integer",
    }
}

PATH_PARAM_Y = {
    "name": "y",
    "in": "path",
    "description": "The tile grid's y-coordinate",
    "schema": {
        "type": "integer",
    }
}

PATH_PARAM_Z = {
    "name": "z",
    "in": "path",
    "description": "The tile grid's z-coordinate",
    "schema": {
        "type": "integer",
    }
}

QUERY_PARAM_TIME = {
    "name": "time",
    "in": "query",
    "description": "Optional time coordinate using format"
                   " \"YYYY-MM-DD hh:mm:ss\"",
    "schema": {
        "type": "string",
        "format": "datetime"
    }
}

QUERY_PARAM_FORMAT = {
    "name": "format",
    "in": "query",
    "description": "Image format",
    "schema": {
        "type": "string",
        "enum": ["png", "image/png"],
        "default": "png"
    }
}

QUERY_PARAM_RETINA = {
    "name": "retina",
    "in": "query",
    "description": "Returns tiles of size"
                   " 512 instead of 256",
    "schema": {
        "type": "boolean"
    }
}

TILE_PARAMETERS = [
    PATH_PARAM_DATASET_ID,
    PATH_PARAM_VAR_NAME,
    PATH_PARAM_Z,
    PATH_PARAM_Y,
    PATH_PARAM_X,
    QUERY_PARAM_CRS,
    QUERY_PARAM_VMIN,
    QUERY_PARAM_VMAX,
    QUERY_PARAM_CBAR,
    QUERY_PARAM_TIME,
    QUERY_PARAM_FORMAT,
    QUERY_PARAM_RETINA,
]


# noinspection PyPep8Naming
class TilesHandler(ApiHandler[TilesContext]):
    async def get(self,
                  datasetId: str,
                  varName: str,
                  z: str, y: str, x: str):
        tile = await self.ctx.run_in_executor(
            None,
            compute_ml_dataset_tile,
            self.ctx,
            datasetId,
            varName,
            None,
            x, y, z,
            {k: v[0] for k, v in self.request.query.items()}
        )
        self.response.set_header('Content-Type', 'image/png')
        await self.response.finish(tile)


# noinspection PyPep8Naming
@api.route('/tiles/{datasetId}/{varName}/{z}/{y}/{x}')
class NewTilesHandler(TilesHandler):
    @api.operation(operation_id='getTile',
                   summary="Get the image tile for a variable"
                           " and given tile grid coordinates.",
                   parameters=TILE_PARAMETERS)
    async def get(self,
                  datasetId: str,
                  varName: str,
                  z: str, y: str, x: str):
        await super().get(datasetId, varName, z, y, x)


# For backward compatibility only
# noinspection PyPep8Naming
@api.route('/datasets/{datasetId}/vars/{varName}/tiles2/{z}/{y}/{x}')
class OldTilesHandler(TilesHandler):
    """Deprecated. Use /tiles/{datasetId}/{varName}/{z}/{y}/{x}."""
    @api.operation(operation_id='getDatasetVariableTile',
                   tags=['datasets'],
                   summary="Get the image tile for a variable"
                           " and given tile grid coordinates. (deprecated)",
                   parameters=TILE_PARAMETERS)
    async def get(self,
                  datasetId: str,
                  varName: str,
                  z: str, y: str, x: str):
        await super().get(datasetId, varName, z, y, x)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math
from fractions import Fraction
from typing import Any
from typing import Tuple, Optional, Union

import affine
import dask.array as da
import numpy as np
import pyproj.crs
import xarray as xr

from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.undefined import UNDEFINED

Number = Union[int, float]
AffineTransformMatrix = Tuple[Tuple[Number, Number, Number],
                              Tuple[Number, Number, Number]]


def _to_int_or_float(x: Number) -> Number:
    """
    If x is an int or is close to an int return it
    as int otherwise as float. Helps avoiding errors
    introduced by inaccurate floating point ops.
    """
    if isinstance(x, int):
        return x
    xf = float(x)
    xi = round(xf)
    return xi if math.isclose(xi, xf, rel_tol=1e-5) else xf


def _from_affine(matrix: affine.Affine) -> AffineTransformMatrix:
    return (matrix.a, matrix.b, matrix.c), (matrix.d, matrix.e, matrix.f)


def _to_affine(matrix: AffineTransformMatrix) -> affine.Affine:
    return affine.Affine(*matrix[0], *matrix[1])


def _normalize_crs(crs: Union[str, pyproj.CRS]) -> pyproj.CRS:
    if isinstance(crs, pyproj.CRS):
        return crs
    assert_instance(crs, str, 'crs')
    return pyproj.CRS.from_string(crs)


def _normalize_int_pair(
        value: Any,
        name: str = None,
        default: Optional[Tuple[int, int]] = UNDEFINED
) -> Optional[Tuple[int, int]]:
    if isinstance(value, int):
        return value, value
    elif value is not None:
        x, y = value
        return int(x), int(y)
    elif default != UNDEFINED:
        return default
    else:
        assert_given(name, 'name')
        raise ValueError(f'{name} must be an int'
                         f' or a sequence of two ints')


def _normalize_number_pair(
        value: Any,
        name: str = None,
        default: Optional[Tuple[Number, Number]] = UNDEFINED
) -> Optional[Tuple[Number, Number]]:
    if isinstance(value, (float, int)):
        x, y = value, value
        return _to_int_or_float(x), _to_int_or_float(y)
    elif value is not None:
        x, y = value
        return _to_int_or_float(x), _to_int_or_float(y)
    elif default != UNDEFINED:
        return default
    else:
        assert_given(name, 'name')
        raise ValueError(f'{name} must be a number'
                         f' or a sequence of two numbers')


def to_lon_360(lon_var: Union[np.ndarray, da.Array, xr.DataArray]):
    if isinstance(lon_var, xr.DataArray):
        return lon_var.where(lon_var >= 0.0, lon_var + 360.0)
    else:
        lon_var = da.asarray(lon_var)
        return da.where(lon_var >= 0.0, lon_var, lon_var + 360.0)


def from_lon_360(lon_var: Union[np.ndarray, da.Array, xr.DataArray]):
    if isinstance(lon_var, xr.DataArray):
        return lon_var.where(lon_var <= 180.0, lon_var - 360.0)
    else:
        lon_var = da.asarray(lon_var)
        return da.where(lon_var <= 180.0, lon_var, lon_var - 360.0)


def _default_xy_var_names(crs: pyproj.crs.CRS) -> Tuple[str, str]:
    return ('lon', 'lat') if crs.is_geographic else ('x', 'y')


def _default_xy_dim_names(crs: pyproj.crs.CRS) -> Tuple[str, str]:
    return _default_xy_var_names(crs)


def _assert_valid_xy_names(value: Any, name: str = None):
    assert_instance(value, tuple, name=name)
    assert_true(len(value) == 2
                and all(value)
                and value[0] != value[1],
                f'invalid {name or "value"}')


def _assert_valid_xy_coords(xy_coords: Any):
    assert_instance(xy_coords, xr.DataArray, name='xy_coords')
    assert_true(xy_coords.ndim == 3
                and xy_coords.shape[0] == 2
                and xy_coords.shape[1] >= 2
                and xy_coords.shape[2] >= 2,
                'xy_coords must have dimensions'
                ' (2, height, width) with height >= 2 and width >= 2')


_RESOLUTIONS = {
    10: (1, 0),
    20: (2, 0),
    25: (25, 1),
    50: (5, 0),
    100: (1, -1),
}

_RESOLUTION_SET = set(k / 100 for k in _RESOLUTIONS.keys())


def round_to_fraction(value: float,
                      digits: int = 2,
                      resolution: float = 1) -> Fraction:
    """
    Round *value* at position given by significant
    *digits* and return result as fraction.

    :param value: The value
    :param digits: The number of significant digits.
        Must be an integer >= 1. Default is 2.
    :param resolution: The rounding resolution
        for the least significant digit.
        Must be one of (0.1, 0.2, 0.25, 0.5, 1). Default is 1.
    :return: The rounded value as fraction.Fraction instance.
    """
    if digits < 1:
        raise ValueError('digits must be a positive integer')
    resolution_key = round(100 * resolution)
    if resolution_key not in _RESOLUTIONS \
            or not math.isclose(100 * resolution, resolution_key):
        raise ValueError(f'resolution must be one of {_RESOLUTION_SET}')
    if value == 0:
        return Fraction(0, 1)
    sign = 1
    if value < 0:
        sign = -1
        value = -value
    resolution, resolution_digits = _RESOLUTIONS[resolution_key]
    exponent = math.floor(math.log10(value)) - digits - resolution_digits
    if exponent >= 0:
        magnitude = Fraction(10 ** exponent, 1)
    else:
        magnitude = Fraction(1, 10 ** -exponent)
    scaled_value = value / magnitude
    discrete_value = resolution * round(scaled_value / resolution)
    return (sign * discrete_value) * magnitude


def scale_xy_res_and_size(xy_res: Tuple[float, float],
                          size: Tuple[int, int],
                          xy_scale: Tuple[float, float]) \
        -> Tuple[Tuple[float, float], Tuple[int, int]]:
    """
    Scale given *xy_res* and *size* using *xy_scale*.
    Make sure, size components are not less than 2.
    """
    x_res, y_res = xy_res
    x_scale, y_scale = xy_scale
    w, h = size
    w, h = round(x_scale * w), round(y_scale * h)
    return ((x_res / x_scale,
             y_res / y_scale),
            (w if w >= 2 else 2,
             h if h >= 2 else 2))
